from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.create_instance_request_features_item import (
    CreateInstanceRequestFeaturesItem,
    check_create_instance_request_features_item,
)
from ..models.create_instance_request_restart_policy import (
    CreateInstanceRequestRestartPolicy,
    check_create_instance_request_restart_policy,
)
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.create_instance_request_env import CreateInstanceRequestEnv
    from ..models.create_instance_request_service_group import CreateInstanceRequestServiceGroup
    from ..models.create_instance_request_volume import CreateInstanceRequestVolume
    from ..models.instance_scale_to_zero import InstanceScaleToZero


T = TypeVar("T", bound="CreateInstanceRequest")


@_attrs_define
class CreateInstanceRequest:
    """The request message for creating a new instance."""

    image: str
    """ The image to use for the instance. """
    name: Union[Unset, str] = UNSET
    """ (Optional).  The name of the instance.

    If not provided, a random name will be generated.  The name must be unique. """
    args: Union[Unset, list[str]] = UNSET
    """ (Optional).  The arguments to pass to the instance when it starts. """
    env: Union[Unset, "CreateInstanceRequestEnv"] = UNSET
    """ (Optional).  Environment variables to set for the instance. """
    memory_mb: Union[Unset, int] = UNSET
    """ (Optional).  Memory in MB to allocate for the instance.  Default is 128. """
    service_group: Union[Unset, "CreateInstanceRequestServiceGroup"] = UNSET
    """ The service group configuration when creating an instance.

    If no existing (persistent) service group is specified via its identifier,
    a new (ephemeral) service group can be created by specifying the services
    it should expose.  A service defines the configuration settings of an
    exposed port by the instance.  A service is a combination of a public port,
    an internal port, and a set of handlers that define how the service will
    handle incoming connections. """
    volumes: Union[Unset, list["CreateInstanceRequestVolume"]] = UNSET
    """ Volumes to attach to the instance.

    This list can contain both existing and new volumes to create as part of
    the instance creation.  Existing volumes can be referenced by their name or
    UUID.  New volumes can be created by specifying a name, size in MiB, and
    mount point in the instance.  The mount point is the directory in the
    instance where the volume will be mounted. """
    autostart: Union[Unset, bool] = UNSET
    """ Whether the instance should start automatically on creation. """
    replicas: Union[Unset, int] = UNSET
    """ Number of replicas for the instance. """
    restart_policy: Union[Unset, CreateInstanceRequestRestartPolicy] = UNSET
    """ Restart policy for the instance.  This defines how the instance should
    behave when it stops or crashes. """
    scale_to_zero: Union[Unset, "InstanceScaleToZero"] = UNSET
    """ Scale-to-zero defines the configuration for scaling the instance to zero.
    When an instance is scaled-to-zero it can be either stopped (and fully
    shutdown) or paused wherein the state of the instance is preserved (e.g., RAM
    contents) and the instance can be resumed later without losing its state,
    i.e. "stateful". """
    vcpus: Union[Unset, int] = UNSET
    """ Number of vCPUs to allocate for the instance. """
    wait_timeout_ms: Union[Unset, int] = UNSET
    """ Timeout to wait for all new instances to reach running state in
    milliseconds.  If you autostart your new instance, you can wait for it to
    finish starting with a blocking API call if you specify a wait timeout
    greater than zero.  No wait performed for a value of 0. """
    features: Union[Unset, list[CreateInstanceRequestFeaturesItem]] = UNSET
    """ Features to enable for the instance.  Features are specific
    configurations or capabilities that can be enabled for the instance. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        image = self.image

        name = self.name

        args: Union[Unset, list[str]] = UNSET
        if not isinstance(self.args, Unset):
            args = self.args

        env: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.env, Unset):
            env = self.env.to_dict()

        memory_mb = self.memory_mb

        service_group: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.service_group, Unset):
            service_group = self.service_group.to_dict()

        volumes: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.volumes, Unset):
            volumes = []
            for volumes_item_data in self.volumes:
                volumes_item = volumes_item_data.to_dict()
                volumes.append(volumes_item)

        autostart = self.autostart

        replicas = self.replicas

        restart_policy: Union[Unset, str] = UNSET
        if not isinstance(self.restart_policy, Unset):
            restart_policy = self.restart_policy

        scale_to_zero: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.scale_to_zero, Unset):
            scale_to_zero = self.scale_to_zero.to_dict()

        vcpus = self.vcpus

        wait_timeout_ms = self.wait_timeout_ms

        features: Union[Unset, list[str]] = UNSET
        if not isinstance(self.features, Unset):
            features = []
            for features_item_data in self.features:
                features_item: str = features_item_data
                features.append(features_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "image": image,
            }
        )
        if name is not UNSET:
            field_dict["name"] = name
        if args is not UNSET:
            field_dict["args"] = args
        if env is not UNSET:
            field_dict["env"] = env
        if memory_mb is not UNSET:
            field_dict["memory_mb"] = memory_mb
        if service_group is not UNSET:
            field_dict["service_group"] = service_group
        if volumes is not UNSET:
            field_dict["volumes"] = volumes
        if autostart is not UNSET:
            field_dict["autostart"] = autostart
        if replicas is not UNSET:
            field_dict["replicas"] = replicas
        if restart_policy is not UNSET:
            field_dict["restart_policy"] = restart_policy
        if scale_to_zero is not UNSET:
            field_dict["scale_to_zero"] = scale_to_zero
        if vcpus is not UNSET:
            field_dict["vcpus"] = vcpus
        if wait_timeout_ms is not UNSET:
            field_dict["wait_timeout_ms"] = wait_timeout_ms
        if features is not UNSET:
            field_dict["features"] = features

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.create_instance_request_env import CreateInstanceRequestEnv
        from ..models.create_instance_request_service_group import CreateInstanceRequestServiceGroup
        from ..models.create_instance_request_volume import CreateInstanceRequestVolume
        from ..models.instance_scale_to_zero import InstanceScaleToZero

        d = dict(src_dict)
        image = d.pop("image")

        name = d.pop("name", UNSET)

        args = cast(list[str], d.pop("args", UNSET))

        _env = d.pop("env", UNSET)
        env: Union[Unset, CreateInstanceRequestEnv]
        if isinstance(_env, Unset):
            env = UNSET
        else:
            env = CreateInstanceRequestEnv.from_dict(_env)

        memory_mb = d.pop("memory_mb", UNSET)

        _service_group = d.pop("service_group", UNSET)
        service_group: Union[Unset, CreateInstanceRequestServiceGroup]
        if isinstance(_service_group, Unset):
            service_group = UNSET
        else:
            service_group = CreateInstanceRequestServiceGroup.from_dict(_service_group)

        volumes = []
        _volumes = d.pop("volumes", UNSET)
        for volumes_item_data in _volumes or []:
            volumes_item = CreateInstanceRequestVolume.from_dict(volumes_item_data)

            volumes.append(volumes_item)

        autostart = d.pop("autostart", UNSET)

        replicas = d.pop("replicas", UNSET)

        _restart_policy = d.pop("restart_policy", UNSET)
        restart_policy: Union[Unset, CreateInstanceRequestRestartPolicy]
        if isinstance(_restart_policy, Unset):
            restart_policy = UNSET
        else:
            restart_policy = check_create_instance_request_restart_policy(_restart_policy)

        _scale_to_zero = d.pop("scale_to_zero", UNSET)
        scale_to_zero: Union[Unset, InstanceScaleToZero]
        if isinstance(_scale_to_zero, Unset):
            scale_to_zero = UNSET
        else:
            scale_to_zero = InstanceScaleToZero.from_dict(_scale_to_zero)

        vcpus = d.pop("vcpus", UNSET)

        wait_timeout_ms = d.pop("wait_timeout_ms", UNSET)

        features = []
        _features = d.pop("features", UNSET)
        for features_item_data in _features or []:
            features_item = check_create_instance_request_features_item(features_item_data)

            features.append(features_item)

        create_instance_request = cls(
            image=image,
            name=name,
            args=args,
            env=env,
            memory_mb=memory_mb,
            service_group=service_group,
            volumes=volumes,
            autostart=autostart,
            replicas=replicas,
            restart_policy=restart_policy,
            scale_to_zero=scale_to_zero,
            vcpus=vcpus,
            wait_timeout_ms=wait_timeout_ms,
            features=features,
        )

        create_instance_request.additional_properties = d
        return create_instance_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Optional, Union

import httpx

from ...client import AuthenticatedClient, Client
from ...models.attach_volumes_request import AttachVolumesRequest
from ...models.attach_volumes_response import AttachVolumesResponse
from ...types import Response


def _get_kwargs(
    *,
    body: AttachVolumesRequest,
) -> dict[str, Any]:
    headers: dict[str, Any] = {}

    _kwargs: dict[str, Any] = {
        "method": "put",
        "url": "/v1/volumes/attach",
    }

    _kwargs["json"] = body.to_dict()

    headers["Content-Type"] = "application/json"

    _kwargs["headers"] = headers
    return _kwargs


def _parse_response(*, client: Union[AuthenticatedClient, Client], response: httpx.Response) -> AttachVolumesResponse:
    response_default = AttachVolumesResponse.from_dict(response.json())

    return response_default


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[AttachVolumesResponse]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: Union[AuthenticatedClient, Client],
    body: AttachVolumesRequest,
) -> Response[AttachVolumesResponse]:
    """Attach Volumes

     Attach one or more volumes specified by ID(s) (name or UUID) to instances
    so that the volumes are mounted when the instances start.  The volumes need
    to be in `available` state and the instances must be in `stopped` state.

    Args:
        body (AttachVolumesRequest): The request message for attaching one or more volume(s) to
            instances by
            their UUID(s) or name(s).

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[AttachVolumesResponse]
    """

    kwargs = _get_kwargs(
        body=body,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: Union[AuthenticatedClient, Client],
    body: AttachVolumesRequest,
) -> Optional[AttachVolumesResponse]:
    """Attach Volumes

     Attach one or more volumes specified by ID(s) (name or UUID) to instances
    so that the volumes are mounted when the instances start.  The volumes need
    to be in `available` state and the instances must be in `stopped` state.

    Args:
        body (AttachVolumesRequest): The request message for attaching one or more volume(s) to
            instances by
            their UUID(s) or name(s).

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        AttachVolumesResponse
    """

    return sync_detailed(
        client=client,
        body=body,
    ).parsed


async def asyncio_detailed(
    *,
    client: Union[AuthenticatedClient, Client],
    body: AttachVolumesRequest,
) -> Response[AttachVolumesResponse]:
    """Attach Volumes

     Attach one or more volumes specified by ID(s) (name or UUID) to instances
    so that the volumes are mounted when the instances start.  The volumes need
    to be in `available` state and the instances must be in `stopped` state.

    Args:
        body (AttachVolumesRequest): The request message for attaching one or more volume(s) to
            instances by
            their UUID(s) or name(s).

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[AttachVolumesResponse]
    """

    kwargs = _get_kwargs(
        body=body,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: Union[AuthenticatedClient, Client],
    body: AttachVolumesRequest,
) -> Optional[AttachVolumesResponse]:
    """Attach Volumes

     Attach one or more volumes specified by ID(s) (name or UUID) to instances
    so that the volumes are mounted when the instances start.  The volumes need
    to be in `available` state and the instances must be in `stopped` state.

    Args:
        body (AttachVolumesRequest): The request message for attaching one or more volume(s) to
            instances by
            their UUID(s) or name(s).

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        AttachVolumesResponse
    """

    return (
        await asyncio_detailed(
            client=client,
            body=body,
        )
    ).parsed



from typing import Literal, cast

GetInstanceLogsResponseLoggedInstanceState = Literal[
    "draining", "running", "standby", "starting", "stopped", "stopping"
]

GET_INSTANCE_LOGS_RESPONSE_LOGGED_INSTANCE_STATE_VALUES: set[GetInstanceLogsResponseLoggedInstanceState] = {
    "draining",
    "running",
    "standby",
    "starting",
    "stopped",
    "stopping",
}


def check_get_instance_logs_response_logged_instance_state(value: str) -> GetInstanceLogsResponseLoggedInstanceState:
    if value in GET_INSTANCE_LOGS_RESPONSE_LOGGED_INSTANCE_STATE_VALUES:
        return cast(GetInstanceLogsResponseLoggedInstanceState, value)
    raise TypeError(
        f"Unexpected value {value!r}. Expected one of {GET_INSTANCE_LOGS_RESPONSE_LOGGED_INSTANCE_STATE_VALUES!r}"
    )



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.body_instance_id import BodyInstanceID


T = TypeVar("T", bound="AttachVolumeByUUIDRequestBody")


@_attrs_define
class AttachVolumeByUUIDRequestBody:
    attach_to: "BodyInstanceID"
    """ Reference to the instance to detach the volume from. """
    at: str
    """ Path of the mountpoint.

    The path must be absolute, not contain `.` and `..` components, and not
    contain colons (`:`). The path must point to an empty directory. If the
    directory does not exist, it is created. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        attach_to = self.attach_to.to_dict()

        at = self.at

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "attach_to": attach_to,
                "at": at,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.body_instance_id import BodyInstanceID

        d = dict(src_dict)
        attach_to = BodyInstanceID.from_dict(d.pop("attach_to"))

        at = d.pop("at")

        attach_volume_by_uuid_request_body = cls(
            attach_to=attach_to,
            at=at,
        )

        attach_volume_by_uuid_request_body.additional_properties = d
        return attach_volume_by_uuid_request_body

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.name_or_uuid import NameOrUUID


T = TypeVar("T", bound="CreateServiceGroupRequestDomain")


@_attrs_define
class CreateServiceGroupRequestDomain:
    """A domain name"""

    name: str
    """ Publicly accessible domain name.  If this name ends in a period `.` it must
    be a valid Full Qualified Domain Name (FQDN), otherwise it will become a
    subdomain of the target metro. """
    certificate: Union[Unset, "NameOrUUID"] = UNSET
    """ An identifier for a resource.  Either a name or a UUID. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        name = self.name

        certificate: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.certificate, Unset):
            certificate = self.certificate.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
            }
        )
        if certificate is not UNSET:
            field_dict["certificate"] = certificate

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.name_or_uuid import NameOrUUID

        d = dict(src_dict)
        name = d.pop("name")

        _certificate = d.pop("certificate", UNSET)
        certificate: Union[Unset, NameOrUUID]
        if isinstance(_certificate, Unset):
            certificate = UNSET
        else:
            certificate = NameOrUUID.from_dict(_certificate)

        create_service_group_request_domain = cls(
            name=name,
            certificate=certificate,
        )

        create_service_group_request_domain.additional_properties = d
        return create_service_group_request_domain

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.image_labels import ImageLabels
    from ..models.object_ import Object


T = TypeVar("T", bound="Image")


@_attrs_define
class Image:
    tag: Union[Unset, str] = UNSET
    """ The canonical name of the image is known as the "tag". """
    digest: Union[Unset, str] = UNSET
    """ The digest of the image is a unique identifier of the image manifest which
    is a string representation including the hashing algorithm and the hash
    value separated by a colon. """
    description: Union[Unset, str] = UNSET
    """ A description of the image. """
    created_at: Union[Unset, datetime.datetime] = UNSET
    """ When the image was created. """
    arch: Union[Unset, str] = UNSET
    """ The architecture of the image. """
    entrypoint: Union[Unset, list[str]] = UNSET
    """ The entrypoint of the image is the command that is run when the image is
    started. """
    cmd: Union[Unset, list[str]] = UNSET
    """ The command to run when the image is started. """
    env: Union[Unset, list[str]] = UNSET
    """ The environment variables to set when the image is started. """
    ports: Union[Unset, list[str]] = UNSET
    """ Documented port mappings for the image. """
    volumes: Union[Unset, list[str]] = UNSET
    """ Documented volumes for the image. """
    labels: Union[Unset, "ImageLabels"] = UNSET
    """ Labels are key-value pairs. """
    workdir: Union[Unset, str] = UNSET
    """ The working directory for the image is the directory that is set as the
    current working directory when the image is started. """
    kernel: Union[Unset, "Object"] = UNSET
    """ An object is a single component of an image which is external and can be
    uniquely identified by its digest. """
    auxiliary_roms: Union[Unset, list["Object"]] = UNSET
    """ List of auxiliary ROMs that are used by the image. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        tag = self.tag

        digest = self.digest

        description = self.description

        created_at: Union[Unset, str] = UNSET
        if not isinstance(self.created_at, Unset):
            created_at = self.created_at.isoformat()

        arch = self.arch

        entrypoint: Union[Unset, list[str]] = UNSET
        if not isinstance(self.entrypoint, Unset):
            entrypoint = self.entrypoint

        cmd: Union[Unset, list[str]] = UNSET
        if not isinstance(self.cmd, Unset):
            cmd = self.cmd

        env: Union[Unset, list[str]] = UNSET
        if not isinstance(self.env, Unset):
            env = self.env

        ports: Union[Unset, list[str]] = UNSET
        if not isinstance(self.ports, Unset):
            ports = self.ports

        volumes: Union[Unset, list[str]] = UNSET
        if not isinstance(self.volumes, Unset):
            volumes = self.volumes

        labels: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.labels, Unset):
            labels = self.labels.to_dict()

        workdir = self.workdir

        kernel: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.kernel, Unset):
            kernel = self.kernel.to_dict()

        auxiliary_roms: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.auxiliary_roms, Unset):
            auxiliary_roms = []
            for auxiliary_roms_item_data in self.auxiliary_roms:
                auxiliary_roms_item = auxiliary_roms_item_data.to_dict()
                auxiliary_roms.append(auxiliary_roms_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if tag is not UNSET:
            field_dict["tag"] = tag
        if digest is not UNSET:
            field_dict["digest"] = digest
        if description is not UNSET:
            field_dict["description"] = description
        if created_at is not UNSET:
            field_dict["created_at"] = created_at
        if arch is not UNSET:
            field_dict["arch"] = arch
        if entrypoint is not UNSET:
            field_dict["entrypoint"] = entrypoint
        if cmd is not UNSET:
            field_dict["cmd"] = cmd
        if env is not UNSET:
            field_dict["env"] = env
        if ports is not UNSET:
            field_dict["ports"] = ports
        if volumes is not UNSET:
            field_dict["volumes"] = volumes
        if labels is not UNSET:
            field_dict["labels"] = labels
        if workdir is not UNSET:
            field_dict["workdir"] = workdir
        if kernel is not UNSET:
            field_dict["kernel"] = kernel
        if auxiliary_roms is not UNSET:
            field_dict["auxiliary_roms"] = auxiliary_roms

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.image_labels import ImageLabels
        from ..models.object_ import Object

        d = dict(src_dict)
        tag = d.pop("tag", UNSET)

        digest = d.pop("digest", UNSET)

        description = d.pop("description", UNSET)

        _created_at = d.pop("created_at", UNSET)
        created_at: Union[Unset, datetime.datetime]
        if isinstance(_created_at, Unset):
            created_at = UNSET
        else:
            created_at = isoparse(_created_at)

        arch = d.pop("arch", UNSET)

        entrypoint = cast(list[str], d.pop("entrypoint", UNSET))

        cmd = cast(list[str], d.pop("cmd", UNSET))

        env = cast(list[str], d.pop("env", UNSET))

        ports = cast(list[str], d.pop("ports", UNSET))

        volumes = cast(list[str], d.pop("volumes", UNSET))

        _labels = d.pop("labels", UNSET)
        labels: Union[Unset, ImageLabels]
        if isinstance(_labels, Unset):
            labels = UNSET
        else:
            labels = ImageLabels.from_dict(_labels)

        workdir = d.pop("workdir", UNSET)

        _kernel = d.pop("kernel", UNSET)
        kernel: Union[Unset, Object]
        if isinstance(_kernel, Unset):
            kernel = UNSET
        else:
            kernel = Object.from_dict(_kernel)

        auxiliary_roms = []
        _auxiliary_roms = d.pop("auxiliary_roms", UNSET)
        for auxiliary_roms_item_data in _auxiliary_roms or []:
            auxiliary_roms_item = Object.from_dict(auxiliary_roms_item_data)

            auxiliary_roms.append(auxiliary_roms_item)

        image = cls(
            tag=tag,
            digest=digest,
            description=description,
            created_at=created_at,
            arch=arch,
            entrypoint=entrypoint,
            cmd=cmd,
            env=env,
            ports=ports,
            volumes=volumes,
            labels=labels,
            workdir=workdir,
            kernel=kernel,
            auxiliary_roms=auxiliary_roms,
        )

        image.additional_properties = d
        return image

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Optional, Union

import httpx

from ...client import AuthenticatedClient, Client
from ...models.detach_volumes_request import DetachVolumesRequest
from ...models.detach_volumes_response import DetachVolumesResponse
from ...types import Response


def _get_kwargs(
    *,
    body: DetachVolumesRequest,
) -> dict[str, Any]:
    headers: dict[str, Any] = {}

    _kwargs: dict[str, Any] = {
        "method": "put",
        "url": "/v1/volumes/detach",
    }

    _kwargs["json"] = body.to_dict()

    headers["Content-Type"] = "application/json"

    _kwargs["headers"] = headers
    return _kwargs


def _parse_response(*, client: Union[AuthenticatedClient, Client], response: httpx.Response) -> DetachVolumesResponse:
    response_default = DetachVolumesResponse.from_dict(response.json())

    return response_default


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[DetachVolumesResponse]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: Union[AuthenticatedClient, Client],
    body: DetachVolumesRequest,
) -> Response[DetachVolumesResponse]:
    """Detach Volumes

     Detach volumes specified by ID(s) (name or UUID) from instances.  If no
    particular instance is specified the volume is detached from all instances.
    The instances from which to detach must not have the volumes mounted.  The
    API returns an error for each instance from which it was unable to detach
    the volume.  If the volume has been created together with an instance,
    detaching the volume will make it persistent (i.e., it survives the
    deletion of the instance).

    Args:
        body (DetachVolumesRequest): The request message for detaching one or more volume(s) from
            instances by
            their UUID(s) or name(s).

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[DetachVolumesResponse]
    """

    kwargs = _get_kwargs(
        body=body,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: Union[AuthenticatedClient, Client],
    body: DetachVolumesRequest,
) -> Optional[DetachVolumesResponse]:
    """Detach Volumes

     Detach volumes specified by ID(s) (name or UUID) from instances.  If no
    particular instance is specified the volume is detached from all instances.
    The instances from which to detach must not have the volumes mounted.  The
    API returns an error for each instance from which it was unable to detach
    the volume.  If the volume has been created together with an instance,
    detaching the volume will make it persistent (i.e., it survives the
    deletion of the instance).

    Args:
        body (DetachVolumesRequest): The request message for detaching one or more volume(s) from
            instances by
            their UUID(s) or name(s).

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        DetachVolumesResponse
    """

    return sync_detailed(
        client=client,
        body=body,
    ).parsed


async def asyncio_detailed(
    *,
    client: Union[AuthenticatedClient, Client],
    body: DetachVolumesRequest,
) -> Response[DetachVolumesResponse]:
    """Detach Volumes

     Detach volumes specified by ID(s) (name or UUID) from instances.  If no
    particular instance is specified the volume is detached from all instances.
    The instances from which to detach must not have the volumes mounted.  The
    API returns an error for each instance from which it was unable to detach
    the volume.  If the volume has been created together with an instance,
    detaching the volume will make it persistent (i.e., it survives the
    deletion of the instance).

    Args:
        body (DetachVolumesRequest): The request message for detaching one or more volume(s) from
            instances by
            their UUID(s) or name(s).

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[DetachVolumesResponse]
    """

    kwargs = _get_kwargs(
        body=body,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: Union[AuthenticatedClient, Client],
    body: DetachVolumesRequest,
) -> Optional[DetachVolumesResponse]:
    """Detach Volumes

     Detach volumes specified by ID(s) (name or UUID) from instances.  If no
    particular instance is specified the volume is detached from all instances.
    The instances from which to detach must not have the volumes mounted.  The
    API returns an error for each instance from which it was unable to detach
    the volume.  If the volume has been created together with an instance,
    detaching the volume will make it persistent (i.e., it survives the
    deletion of the instance).

    Args:
        body (DetachVolumesRequest): The request message for detaching one or more volume(s) from
            instances by
            their UUID(s) or name(s).

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        DetachVolumesResponse
    """

    return (
        await asyncio_detailed(
            client=client,
            body=body,
        )
    ).parsed



from collections.abc import Mapping
from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="InstanceServiceGroupInstanceDomain")


@_attrs_define
class InstanceServiceGroupInstanceDomain:
    """The domain configuration for the service group.

    Domain names are completely specified with all labels in the hierarchy of
    the DNS, having no parts omitted.  The domain can be associated with an
    existing certificate by specifying the certificate's name or UUID.  If no
    certificate is specified and a FQDN is provided, Unikraft Cloud will
    automatically generate a new certificate for the domain based on Let's
    Encrypt and seek to accomplish a DNS-01 challenge.

    """

    fqdn: Union[Unset, str] = UNSET
    """ Publicly accessible domain name.

    If this name ends in a period `.` it must be a valid Full Qualified
    Domain Name (FQDN), otherwise it will become a subdomain of the target
    metro. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        fqdn = self.fqdn

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if fqdn is not UNSET:
            field_dict["fqdn"] = fqdn

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        fqdn = d.pop("fqdn", UNSET)

        instance_service_group_instance_domain = cls(
            fqdn=fqdn,
        )

        instance_service_group_instance_domain.additional_properties = d
        return instance_service_group_instance_domain

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.service_handlers_item import ServiceHandlersItem, check_service_handlers_item
from ..types import UNSET, Unset

T = TypeVar("T", bound="Service")


@_attrs_define
class Service:
    """A service connects a public-facing port to an internal destination port on
    which an application instance listens on.  Additional handlers can be defined
    for each published port in order to define how the service will handle
    incoming connections and forward traffic from the Internet to your
    application.  For example, a service can be configured to terminate TLS
    connections, redirect HTTP traffic, or enable HTTP mode for load balancing.

    """

    port: int
    """ This is the public-facing port that the service will be accessible from
    on the Internet. """
    destination_port: Union[Unset, int] = UNSET
    """ The port number that the instance is listening on.  This is the internal
    port which Unikraft Cloud will forward traffic to. """
    handlers: Union[Unset, list[ServiceHandlersItem]] = UNSET
    """ Connection handlers to use for the service.  Handlers define how the
    service will handle incoming connections and forward traffic from the
    Internet to your application.  For example, a service can be configured
    to terminate TLS connections, redirect HTTP traffic, or enable HTTP mode
    for load balancing.  You configure the handlers for every published
    service port individually. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        port = self.port

        destination_port = self.destination_port

        handlers: Union[Unset, list[str]] = UNSET
        if not isinstance(self.handlers, Unset):
            handlers = []
            for handlers_item_data in self.handlers:
                handlers_item: str = handlers_item_data
                handlers.append(handlers_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "port": port,
            }
        )
        if destination_port is not UNSET:
            field_dict["destination_port"] = destination_port
        if handlers is not UNSET:
            field_dict["handlers"] = handlers

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        port = d.pop("port")

        destination_port = d.pop("destination_port", UNSET)

        handlers = []
        _handlers = d.pop("handlers", UNSET)
        for handlers_item_data in _handlers or []:
            handlers_item = check_service_handlers_item(handlers_item_data)

            handlers.append(handlers_item)

        service = cls(
            port=port,
            destination_port=destination_port,
            handlers=handlers,
        )

        service.additional_properties = d
        return service

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.attach_volumes_request_instance_id import AttachVolumesRequestInstanceID


T = TypeVar("T", bound="AttachVolumesRequest")


@_attrs_define
class AttachVolumesRequest:
    """The request message for attaching one or more volume(s) to instances by
    their UUID(s) or name(s).

    """

    attach_to: "AttachVolumesRequestInstanceID"
    """ Reference to the instance to attach the volume to. """
    at: str
    """ Path of the mountpoint.

    The path must be absolute, not contain `.` and `..` components, and not
    contain colons (`:`). The path must point to an empty directory. If the
    directory does not exist, it is created. """
    uuid: Union[Unset, UUID] = UNSET
    """ The UUID of the volume to attach. Mutually exclusive with name.
    Exactly one of uuid or name must be provided. """
    name: Union[Unset, str] = UNSET
    """ The name of the volume to attach. Mutually exclusive with UUID.
    Exactly one of uuid or name must be provided. """
    readonly: Union[Unset, bool] = UNSET
    """ Whether the volume should be mounted read-only. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        attach_to = self.attach_to.to_dict()

        at = self.at

        uuid: Union[Unset, str] = UNSET
        if not isinstance(self.uuid, Unset):
            uuid = str(self.uuid)

        name = self.name

        readonly = self.readonly

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "attach_to": attach_to,
                "at": at,
            }
        )
        if uuid is not UNSET:
            field_dict["uuid"] = uuid
        if name is not UNSET:
            field_dict["name"] = name
        if readonly is not UNSET:
            field_dict["readonly"] = readonly

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.attach_volumes_request_instance_id import AttachVolumesRequestInstanceID

        d = dict(src_dict)
        attach_to = AttachVolumesRequestInstanceID.from_dict(d.pop("attach_to"))

        at = d.pop("at")

        _uuid = d.pop("uuid", UNSET)
        uuid: Union[Unset, UUID]
        if isinstance(_uuid, Unset):
            uuid = UNSET
        else:
            uuid = UUID(_uuid)

        name = d.pop("name", UNSET)

        readonly = d.pop("readonly", UNSET)

        attach_volumes_request = cls(
            attach_to=attach_to,
            at=at,
            uuid=uuid,
            name=name,
            readonly=readonly,
        )

        attach_volumes_request.additional_properties = d
        return attach_volumes_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Literal, cast

UpdateVolumesRequestItemOp = Literal["add", "del", "set"]

UPDATE_VOLUMES_REQUEST_ITEM_OP_VALUES: set[UpdateVolumesRequestItemOp] = {
    "add",
    "del",
    "set",
}


def check_update_volumes_request_item_op(value: str) -> UpdateVolumesRequestItemOp:
    if value in UPDATE_VOLUMES_REQUEST_ITEM_OP_VALUES:
        return cast(UpdateVolumesRequestItemOp, value)
    raise TypeError(f"Unexpected value {value!r}. Expected one of {UPDATE_VOLUMES_REQUEST_ITEM_OP_VALUES!r}")



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.get_instance_logs_response_logged_instance_state import (
    GetInstanceLogsResponseLoggedInstanceState,
    check_get_instance_logs_response_logged_instance_state,
)
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.get_instance_logs_response_available import GetInstanceLogsResponseAvailable
    from ..models.get_instance_logs_response_range import GetInstanceLogsResponseRange


T = TypeVar("T", bound="GetInstanceLogsResponseLoggedInstance")


@_attrs_define
class GetInstanceLogsResponseLoggedInstance:
    uuid: Union[Unset, UUID] = UNSET
    """ The UUID of the instance. """
    name: Union[Unset, str] = UNSET
    """ The name of the instance. """
    output: Union[Unset, str] = UNSET
    """ Base64 encoded log output of the instance. """
    available: Union[Unset, "GetInstanceLogsResponseAvailable"] = UNSET
    range_: Union[Unset, "GetInstanceLogsResponseRange"] = UNSET
    state: Union[Unset, GetInstanceLogsResponseLoggedInstanceState] = UNSET
    """ State of the instance when the logs were retrieved. """
    message: Union[Unset, str] = UNSET
    """ An optional message providing additional information about the status.
    This field is useful when the status is not `success`. """
    error: Union[Unset, int] = UNSET
    """ An optional error code providing additional information about the status.
    This field is useful when the status is not `success`. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        uuid: Union[Unset, str] = UNSET
        if not isinstance(self.uuid, Unset):
            uuid = str(self.uuid)

        name = self.name

        output = self.output

        available: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.available, Unset):
            available = self.available.to_dict()

        range_: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.range_, Unset):
            range_ = self.range_.to_dict()

        state: Union[Unset, str] = UNSET
        if not isinstance(self.state, Unset):
            state = self.state

        message = self.message

        error = self.error

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if uuid is not UNSET:
            field_dict["uuid"] = uuid
        if name is not UNSET:
            field_dict["name"] = name
        if output is not UNSET:
            field_dict["output"] = output
        if available is not UNSET:
            field_dict["available"] = available
        if range_ is not UNSET:
            field_dict["range"] = range_
        if state is not UNSET:
            field_dict["state"] = state
        if message is not UNSET:
            field_dict["message"] = message
        if error is not UNSET:
            field_dict["error"] = error

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.get_instance_logs_response_available import GetInstanceLogsResponseAvailable
        from ..models.get_instance_logs_response_range import GetInstanceLogsResponseRange

        d = dict(src_dict)
        _uuid = d.pop("uuid", UNSET)
        uuid: Union[Unset, UUID]
        if isinstance(_uuid, Unset):
            uuid = UNSET
        else:
            uuid = UUID(_uuid)

        name = d.pop("name", UNSET)

        output = d.pop("output", UNSET)

        _available = d.pop("available", UNSET)
        available: Union[Unset, GetInstanceLogsResponseAvailable]
        if isinstance(_available, Unset):
            available = UNSET
        else:
            available = GetInstanceLogsResponseAvailable.from_dict(_available)

        _range_ = d.pop("range", UNSET)
        range_: Union[Unset, GetInstanceLogsResponseRange]
        if isinstance(_range_, Unset):
            range_ = UNSET
        else:
            range_ = GetInstanceLogsResponseRange.from_dict(_range_)

        _state = d.pop("state", UNSET)
        state: Union[Unset, GetInstanceLogsResponseLoggedInstanceState]
        if isinstance(_state, Unset):
            state = UNSET
        else:
            state = check_get_instance_logs_response_logged_instance_state(_state)

        message = d.pop("message", UNSET)

        error = d.pop("error", UNSET)

        get_instance_logs_response_logged_instance = cls(
            uuid=uuid,
            name=name,
            output=output,
            available=available,
            range_=range_,
            state=state,
            message=message,
            error=error,
        )

        get_instance_logs_response_logged_instance.additional_properties = d
        return get_instance_logs_response_logged_instance

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.autoscale_policy import AutoscalePolicy
    from ..models.create_autoscale_configuration_by_service_group_uuid_request_instance_create_args import (
        CreateAutoscaleConfigurationByServiceGroupUUIDRequestInstanceCreateArgs,
    )


T = TypeVar("T", bound="CreateAutoscaleConfigurationByServiceGroupUUIDRequest")


@_attrs_define
class CreateAutoscaleConfigurationByServiceGroupUUIDRequest:
    """The request message to create an autoscale configuration for a service group
    based on its UUID.

    """

    uuid: Union[Unset, UUID] = UNSET
    """ The UUID of the service to create a configuration for.
    Mutually exclusive with name. """
    min_size: Union[Unset, int] = UNSET
    """ The minimum number of instances to keep running. """
    max_size: Union[Unset, int] = UNSET
    """ The maximum number of instances to keep running. """
    warmup_time_ms: Union[Unset, int] = UNSET
    """ The warmup time in milliseconds for new instances. """
    cooldown_time_ms: Union[Unset, int] = UNSET
    """ The cooldown time in milliseconds for the autoscale configuration. """
    create_args: Union[Unset, "CreateAutoscaleConfigurationByServiceGroupUUIDRequestInstanceCreateArgs"] = UNSET
    policies: Union[Unset, list["AutoscalePolicy"]] = UNSET
    """ The policies to apply to the autoscale configuration. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        uuid: Union[Unset, str] = UNSET
        if not isinstance(self.uuid, Unset):
            uuid = str(self.uuid)

        min_size = self.min_size

        max_size = self.max_size

        warmup_time_ms = self.warmup_time_ms

        cooldown_time_ms = self.cooldown_time_ms

        create_args: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.create_args, Unset):
            create_args = self.create_args.to_dict()

        policies: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.policies, Unset):
            policies = []
            for policies_item_data in self.policies:
                policies_item = policies_item_data.to_dict()
                policies.append(policies_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if uuid is not UNSET:
            field_dict["uuid"] = uuid
        if min_size is not UNSET:
            field_dict["min_size"] = min_size
        if max_size is not UNSET:
            field_dict["max_size"] = max_size
        if warmup_time_ms is not UNSET:
            field_dict["warmup_time_ms"] = warmup_time_ms
        if cooldown_time_ms is not UNSET:
            field_dict["cooldown_time_ms"] = cooldown_time_ms
        if create_args is not UNSET:
            field_dict["create_args"] = create_args
        if policies is not UNSET:
            field_dict["policies"] = policies

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.autoscale_policy import AutoscalePolicy
        from ..models.create_autoscale_configuration_by_service_group_uuid_request_instance_create_args import (
            CreateAutoscaleConfigurationByServiceGroupUUIDRequestInstanceCreateArgs,
        )

        d = dict(src_dict)
        _uuid = d.pop("uuid", UNSET)
        uuid: Union[Unset, UUID]
        if isinstance(_uuid, Unset):
            uuid = UNSET
        else:
            uuid = UUID(_uuid)

        min_size = d.pop("min_size", UNSET)

        max_size = d.pop("max_size", UNSET)

        warmup_time_ms = d.pop("warmup_time_ms", UNSET)

        cooldown_time_ms = d.pop("cooldown_time_ms", UNSET)

        _create_args = d.pop("create_args", UNSET)
        create_args: Union[Unset, CreateAutoscaleConfigurationByServiceGroupUUIDRequestInstanceCreateArgs]
        if isinstance(_create_args, Unset):
            create_args = UNSET
        else:
            create_args = CreateAutoscaleConfigurationByServiceGroupUUIDRequestInstanceCreateArgs.from_dict(
                _create_args
            )

        policies = []
        _policies = d.pop("policies", UNSET)
        for policies_item_data in _policies or []:
            policies_item = AutoscalePolicy.from_dict(policies_item_data)

            policies.append(policies_item)

        create_autoscale_configuration_by_service_group_uuid_request = cls(
            uuid=uuid,
            min_size=min_size,
            max_size=max_size,
            warmup_time_ms=warmup_time_ms,
            cooldown_time_ms=cooldown_time_ms,
            create_args=create_args,
            policies=policies,
        )

        create_autoscale_configuration_by_service_group_uuid_request.additional_properties = d
        return create_autoscale_configuration_by_service_group_uuid_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, TypeVar, Union
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.stop_instance_response_stopped_instance_previous_state import (
    StopInstanceResponseStoppedInstancePreviousState,
    check_stop_instance_response_stopped_instance_previous_state,
)
from ..models.stop_instance_response_stopped_instance_state import (
    StopInstanceResponseStoppedInstanceState,
    check_stop_instance_response_stopped_instance_state,
)
from ..types import UNSET, Unset

T = TypeVar("T", bound="StopInstanceResponseStoppedInstance")


@_attrs_define
class StopInstanceResponseStoppedInstance:
    uuid: Union[Unset, UUID] = UNSET
    """ The UUID of the instance. """
    name: Union[Unset, str] = UNSET
    """ The name of the instance. """
    state: Union[Unset, StopInstanceResponseStoppedInstanceState] = UNSET
    """ The current state of the instance. """
    previous_state: Union[Unset, StopInstanceResponseStoppedInstancePreviousState] = UNSET
    """ The previous state of the instance before the stop operation was invoked. """
    message: Union[Unset, str] = UNSET
    """ An optional message providing additional information about the status.
    This field is useful when the status is not `success`. """
    error: Union[Unset, int] = UNSET
    """ An optional error code providing additional information about the status.
    This field is useful when the status is not `success`. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        uuid: Union[Unset, str] = UNSET
        if not isinstance(self.uuid, Unset):
            uuid = str(self.uuid)

        name = self.name

        state: Union[Unset, str] = UNSET
        if not isinstance(self.state, Unset):
            state = self.state

        previous_state: Union[Unset, str] = UNSET
        if not isinstance(self.previous_state, Unset):
            previous_state = self.previous_state

        message = self.message

        error = self.error

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if uuid is not UNSET:
            field_dict["uuid"] = uuid
        if name is not UNSET:
            field_dict["name"] = name
        if state is not UNSET:
            field_dict["state"] = state
        if previous_state is not UNSET:
            field_dict["previous_state"] = previous_state
        if message is not UNSET:
            field_dict["message"] = message
        if error is not UNSET:
            field_dict["error"] = error

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        _uuid = d.pop("uuid", UNSET)
        uuid: Union[Unset, UUID]
        if isinstance(_uuid, Unset):
            uuid = UNSET
        else:
            uuid = UUID(_uuid)

        name = d.pop("name", UNSET)

        _state = d.pop("state", UNSET)
        state: Union[Unset, StopInstanceResponseStoppedInstanceState]
        if isinstance(_state, Unset):
            state = UNSET
        else:
            state = check_stop_instance_response_stopped_instance_state(_state)

        _previous_state = d.pop("previous_state", UNSET)
        previous_state: Union[Unset, StopInstanceResponseStoppedInstancePreviousState]
        if isinstance(_previous_state, Unset):
            previous_state = UNSET
        else:
            previous_state = check_stop_instance_response_stopped_instance_previous_state(_previous_state)

        message = d.pop("message", UNSET)

        error = d.pop("error", UNSET)

        stop_instance_response_stopped_instance = cls(
            uuid=uuid,
            name=name,
            state=state,
            previous_state=previous_state,
            message=message,
            error=error,
        )

        stop_instance_response_stopped_instance.additional_properties = d
        return stop_instance_response_stopped_instance

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Literal, cast

UpdateInstancesRequestProp = Literal[
    "args", "delete_lock", "env", "image", "memory_mb", "scale_to_zero", "tags", "vcpus"
]

UPDATE_INSTANCES_REQUEST_PROP_VALUES: set[UpdateInstancesRequestProp] = {
    "args",
    "delete_lock",
    "env",
    "image",
    "memory_mb",
    "scale_to_zero",
    "tags",
    "vcpus",
}


def check_update_instances_request_prop(value: str) -> UpdateInstancesRequestProp:
    if value in UPDATE_INSTANCES_REQUEST_PROP_VALUES:
        return cast(UpdateInstancesRequestProp, value)
    raise TypeError(f"Unexpected value {value!r}. Expected one of {UPDATE_INSTANCES_REQUEST_PROP_VALUES!r}")



from typing import Literal, cast

UpdateServiceGroupsRequestItemProp = Literal["domains", "hard_limit", "services", "soft_limit"]

UPDATE_SERVICE_GROUPS_REQUEST_ITEM_PROP_VALUES: set[UpdateServiceGroupsRequestItemProp] = {
    "domains",
    "hard_limit",
    "services",
    "soft_limit",
}


def check_update_service_groups_request_item_prop(value: str) -> UpdateServiceGroupsRequestItemProp:
    if value in UPDATE_SERVICE_GROUPS_REQUEST_ITEM_PROP_VALUES:
        return cast(UpdateServiceGroupsRequestItemProp, value)
    raise TypeError(f"Unexpected value {value!r}. Expected one of {UPDATE_SERVICE_GROUPS_REQUEST_ITEM_PROP_VALUES!r}")



from collections.abc import Mapping
from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="GetInstanceLogsByUUIDRequestBody")


@_attrs_define
class GetInstanceLogsByUUIDRequestBody:
    offset: Union[Unset, int] = UNSET
    """ The byte offset of the log output to receive.  A negative sign makes the
    offset relative to the end of the log. """
    limit: Union[Unset, int] = UNSET
    """ The amount of bytes to return at most. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        offset = self.offset

        limit = self.limit

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if offset is not UNSET:
            field_dict["offset"] = offset
        if limit is not UNSET:
            field_dict["limit"] = limit

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        offset = d.pop("offset", UNSET)

        limit = d.pop("limit", UNSET)

        get_instance_logs_by_uuid_request_body = cls(
            offset=offset,
            limit=limit,
        )

        get_instance_logs_by_uuid_request_body.additional_properties = d
        return get_instance_logs_by_uuid_request_body

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.create_service_group_request_domain import CreateServiceGroupRequestDomain
    from ..models.service import Service


T = TypeVar("T", bound="CreateServiceGroupRequest")


@_attrs_define
class CreateServiceGroupRequest:
    """The request message for creating a new service group."""

    name: Union[Unset, str] = UNSET
    """ Name of the service group.  This is a human-readable name that can be used
    to identify the service group.  The name must be unique within the context
    of your account.  If no name is specified, a random name is generated for
    you.  The name can also be used to identify the service group in API calls. """
    services: Union[Unset, list["Service"]] = UNSET
    """ Description of exposed services. """
    domains: Union[Unset, list["CreateServiceGroupRequestDomain"]] = UNSET
    """ Description of domains associated with the service group. """
    soft_limit: Union[Unset, int] = UNSET
    """ The soft limit is used by the Unikraft Cloud load balancer to decide when
    to wake up another standby instance.

    For example, if the soft limit is set to 5 and the service consists of 2
    standby instances, one of the instances receives up to 5 concurrent
    requests.  The 6th parallel requests wakes up the second instance.  If
    there are no more standby instances to wake up, the number of requests
    assigned to each instance will exceed the soft limit.  The load balancer
    makes sure that when the number of in-flight requests goes down again,
    instances are put into standby as fast as possible. """
    hard_limit: Union[Unset, int] = UNSET
    """ The hard limit defines the maximum number of concurrent requests that an
    instance assigned to the this service can handle.

    The load balancer will never assign more requests to a single instance.  In
    case there are no other instances available, excess requests fail (i.e.,
    they are blocked and not queued). """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        name = self.name

        services: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.services, Unset):
            services = []
            for services_item_data in self.services:
                services_item = services_item_data.to_dict()
                services.append(services_item)

        domains: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.domains, Unset):
            domains = []
            for domains_item_data in self.domains:
                domains_item = domains_item_data.to_dict()
                domains.append(domains_item)

        soft_limit = self.soft_limit

        hard_limit = self.hard_limit

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if name is not UNSET:
            field_dict["name"] = name
        if services is not UNSET:
            field_dict["services"] = services
        if domains is not UNSET:
            field_dict["domains"] = domains
        if soft_limit is not UNSET:
            field_dict["soft_limit"] = soft_limit
        if hard_limit is not UNSET:
            field_dict["hard_limit"] = hard_limit

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.create_service_group_request_domain import CreateServiceGroupRequestDomain
        from ..models.service import Service

        d = dict(src_dict)
        name = d.pop("name", UNSET)

        services = []
        _services = d.pop("services", UNSET)
        for services_item_data in _services or []:
            services_item = Service.from_dict(services_item_data)

            services.append(services_item)

        domains = []
        _domains = d.pop("domains", UNSET)
        for domains_item_data in _domains or []:
            domains_item = CreateServiceGroupRequestDomain.from_dict(domains_item_data)

            domains.append(domains_item)

        soft_limit = d.pop("soft_limit", UNSET)

        hard_limit = d.pop("hard_limit", UNSET)

        create_service_group_request = cls(
            name=name,
            services=services,
            domains=domains,
            soft_limit=soft_limit,
            hard_limit=hard_limit,
        )

        create_service_group_request.additional_properties = d
        return create_service_group_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.autoscale_policy_adjustment_type import (
    AutoscalePolicyAdjustmentType,
    check_autoscale_policy_adjustment_type,
)
from ..models.autoscale_policy_metric import AutoscalePolicyMetric, check_autoscale_policy_metric
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.autoscale_policy_step import AutoscalePolicyStep


T = TypeVar("T", bound="AutoscalePolicy")


@_attrs_define
class AutoscalePolicy:
    """AutoscalePolicy defines the autoscale policy for a service.
    Right now it contains fields from both the `ondemand` and `step` policies.
    They are marked both as optional, so only one of them should be set at a
    time. This is a current limitation of the API design.

    """

    name: Union[Unset, str] = UNSET
    """ The name of the policy. """
    enabled: Union[Unset, bool] = UNSET
    """ If the policy is enabled. """
    metric: Union[Unset, AutoscalePolicyMetric] = UNSET
    """ Metric to use for the step policy. """
    adjustment_type: Union[Unset, AutoscalePolicyAdjustmentType] = UNSET
    """ The type of adjustment to be made in the step policy. """
    steps: Union[Unset, list["AutoscalePolicyStep"]] = UNSET
    """ The steps for the step policy.
    Each step defines an adjustment value and optional bounds. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        name = self.name

        enabled = self.enabled

        metric: Union[Unset, str] = UNSET
        if not isinstance(self.metric, Unset):
            metric = self.metric

        adjustment_type: Union[Unset, str] = UNSET
        if not isinstance(self.adjustment_type, Unset):
            adjustment_type = self.adjustment_type

        steps: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.steps, Unset):
            steps = []
            for steps_item_data in self.steps:
                steps_item = steps_item_data.to_dict()
                steps.append(steps_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if name is not UNSET:
            field_dict["name"] = name
        if enabled is not UNSET:
            field_dict["enabled"] = enabled
        if metric is not UNSET:
            field_dict["metric"] = metric
        if adjustment_type is not UNSET:
            field_dict["adjustment_type"] = adjustment_type
        if steps is not UNSET:
            field_dict["steps"] = steps

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.autoscale_policy_step import AutoscalePolicyStep

        d = dict(src_dict)
        name = d.pop("name", UNSET)

        enabled = d.pop("enabled", UNSET)

        _metric = d.pop("metric", UNSET)
        metric: Union[Unset, AutoscalePolicyMetric]
        if isinstance(_metric, Unset):
            metric = UNSET
        else:
            metric = check_autoscale_policy_metric(_metric)

        _adjustment_type = d.pop("adjustment_type", UNSET)
        adjustment_type: Union[Unset, AutoscalePolicyAdjustmentType]
        if isinstance(_adjustment_type, Unset):
            adjustment_type = UNSET
        else:
            adjustment_type = check_autoscale_policy_adjustment_type(_adjustment_type)

        steps = []
        _steps = d.pop("steps", UNSET)
        for steps_item_data in _steps or []:
            steps_item = AutoscalePolicyStep.from_dict(steps_item_data)

            steps.append(steps_item)

        autoscale_policy = cls(
            name=name,
            enabled=enabled,
            metric=metric,
            adjustment_type=adjustment_type,
            steps=steps,
        )

        autoscale_policy.additional_properties = d
        return autoscale_policy

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="AutoscalePolicyStep")


@_attrs_define
class AutoscalePolicyStep:
    adjustment: Union[Unset, int] = UNSET
    """ The adjustment value for the step. """
    lower_bound: Union[Unset, int] = UNSET
    """ Lower bound for the step. """
    upper_bound: Union[Unset, int] = UNSET
    """ Upper bound for the step. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        adjustment = self.adjustment

        lower_bound = self.lower_bound

        upper_bound = self.upper_bound

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if adjustment is not UNSET:
            field_dict["adjustment"] = adjustment
        if lower_bound is not UNSET:
            field_dict["lower_bound"] = lower_bound
        if upper_bound is not UNSET:
            field_dict["upper_bound"] = upper_bound

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        adjustment = d.pop("adjustment", UNSET)

        lower_bound = d.pop("lower_bound", UNSET)

        upper_bound = d.pop("upper_bound", UNSET)

        autoscale_policy_step = cls(
            adjustment=adjustment,
            lower_bound=lower_bound,
            upper_bound=upper_bound,
        )

        autoscale_policy_step.additional_properties = d
        return autoscale_policy_step

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Literal, cast

UpdateVolumeByUUIDRequestBodyProp = Literal["delete_lock", "quota_policy", "size_mb", "tags"]

UPDATE_VOLUME_BY_UUID_REQUEST_BODY_PROP_VALUES: set[UpdateVolumeByUUIDRequestBodyProp] = {
    "delete_lock",
    "quota_policy",
    "size_mb",
    "tags",
}


def check_update_volume_by_uuid_request_body_prop(value: str) -> UpdateVolumeByUUIDRequestBodyProp:
    if value in UPDATE_VOLUME_BY_UUID_REQUEST_BODY_PROP_VALUES:
        return cast(UpdateVolumeByUUIDRequestBodyProp, value)
    raise TypeError(f"Unexpected value {value!r}. Expected one of {UPDATE_VOLUME_BY_UUID_REQUEST_BODY_PROP_VALUES!r}")



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.response_status import ResponseStatus, check_response_status
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.autoscale_policy import AutoscalePolicy
    from ..models.service_group_template import ServiceGroupTemplate


T = TypeVar("T", bound="GetAutoscaleConfigurationsResponseServiceGroup")


@_attrs_define
class GetAutoscaleConfigurationsResponseServiceGroup:
    status: Union[Unset, ResponseStatus] = UNSET
    """ The response status of an API request. """
    uuid: Union[Unset, UUID] = UNSET
    """ The UUID of the service where the configuration was created. """
    name: Union[Unset, str] = UNSET
    """ The name of the service where the configuration was created. """
    enabled: Union[Unset, bool] = UNSET
    """ If the autoscale configuration is enabled. """
    min_size: Union[Unset, int] = UNSET
    """ The minimum number of instances to keep running.
    Only if enabled is true. """
    max_size: Union[Unset, int] = UNSET
    """ The maximum number of instances to keep running.
    Only if enabled is true. """
    warmup_time_ms: Union[Unset, int] = UNSET
    """ The warmup time in seconds for new instances.
    Only if enabled is true. """
    cooldown_time_ms: Union[Unset, int] = UNSET
    """ The cooldown time in seconds for the autoscale configuration.
    Only if enabled is true. """
    template: Union[Unset, "ServiceGroupTemplate"] = UNSET
    policies: Union[Unset, list["AutoscalePolicy"]] = UNSET
    """ The policies applied to the autoscale configuration. """
    message: Union[Unset, str] = UNSET
    """ An optional message providing additional information about the status.
    This field is useful when the status is not `success`. """
    error: Union[Unset, int] = UNSET
    """ An optional error code providing additional information about the status.
    This field is useful when the status is not `success`. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        status: Union[Unset, str] = UNSET
        if not isinstance(self.status, Unset):
            status = self.status

        uuid: Union[Unset, str] = UNSET
        if not isinstance(self.uuid, Unset):
            uuid = str(self.uuid)

        name = self.name

        enabled = self.enabled

        min_size = self.min_size

        max_size = self.max_size

        warmup_time_ms = self.warmup_time_ms

        cooldown_time_ms = self.cooldown_time_ms

        template: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.template, Unset):
            template = self.template.to_dict()

        policies: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.policies, Unset):
            policies = []
            for policies_item_data in self.policies:
                policies_item = policies_item_data.to_dict()
                policies.append(policies_item)

        message = self.message

        error = self.error

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if status is not UNSET:
            field_dict["status"] = status
        if uuid is not UNSET:
            field_dict["uuid"] = uuid
        if name is not UNSET:
            field_dict["name"] = name
        if enabled is not UNSET:
            field_dict["enabled"] = enabled
        if min_size is not UNSET:
            field_dict["min_size"] = min_size
        if max_size is not UNSET:
            field_dict["max_size"] = max_size
        if warmup_time_ms is not UNSET:
            field_dict["warmup_time_ms"] = warmup_time_ms
        if cooldown_time_ms is not UNSET:
            field_dict["cooldown_time_ms"] = cooldown_time_ms
        if template is not UNSET:
            field_dict["template"] = template
        if policies is not UNSET:
            field_dict["policies"] = policies
        if message is not UNSET:
            field_dict["message"] = message
        if error is not UNSET:
            field_dict["error"] = error

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.autoscale_policy import AutoscalePolicy
        from ..models.service_group_template import ServiceGroupTemplate

        d = dict(src_dict)
        _status = d.pop("status", UNSET)
        status: Union[Unset, ResponseStatus]
        if isinstance(_status, Unset):
            status = UNSET
        else:
            status = check_response_status(_status)

        _uuid = d.pop("uuid", UNSET)
        uuid: Union[Unset, UUID]
        if isinstance(_uuid, Unset):
            uuid = UNSET
        else:
            uuid = UUID(_uuid)

        name = d.pop("name", UNSET)

        enabled = d.pop("enabled", UNSET)

        min_size = d.pop("min_size", UNSET)

        max_size = d.pop("max_size", UNSET)

        warmup_time_ms = d.pop("warmup_time_ms", UNSET)

        cooldown_time_ms = d.pop("cooldown_time_ms", UNSET)

        _template = d.pop("template", UNSET)
        template: Union[Unset, ServiceGroupTemplate]
        if isinstance(_template, Unset):
            template = UNSET
        else:
            template = ServiceGroupTemplate.from_dict(_template)

        policies = []
        _policies = d.pop("policies", UNSET)
        for policies_item_data in _policies or []:
            policies_item = AutoscalePolicy.from_dict(policies_item_data)

            policies.append(policies_item)

        message = d.pop("message", UNSET)

        error = d.pop("error", UNSET)

        get_autoscale_configurations_response_service_group = cls(
            status=status,
            uuid=uuid,
            name=name,
            enabled=enabled,
            min_size=min_size,
            max_size=max_size,
            warmup_time_ms=warmup_time_ms,
            cooldown_time_ms=cooldown_time_ms,
            template=template,
            policies=policies,
            message=message,
            error=error,
        )

        get_autoscale_configurations_response_service_group.additional_properties = d
        return get_autoscale_configurations_response_service_group

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.create_instance_request_domain import CreateInstanceRequestDomain
    from ..models.service import Service


T = TypeVar("T", bound="CreateInstanceRequestServiceGroup")


@_attrs_define
class CreateInstanceRequestServiceGroup:
    """The service group configuration when creating an instance.

    If no existing (persistent) service group is specified via its identifier,
    a new (ephemeral) service group can be created by specifying the services
    it should expose.  A service defines the configuration settings of an
    exposed port by the instance.  A service is a combination of a public port,
    an internal port, and a set of handlers that define how the service will
    handle incoming connections.

    """

    domains: Union[Unset, list["CreateInstanceRequestDomain"]] = UNSET
    """ Similarly, if no existing (persistent) service group is specified via its
    identifier, a new (ephemeral) service group can be created.  In addition
    to the services it must expose, you can specify which domains it should
    use too. """
    services: Union[Unset, list["Service"]] = UNSET
    """ If no existing service identifier is provided, one or more new
    (ephemeral, non-persistent) service(s) can be created with the following
    definitions. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        domains: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.domains, Unset):
            domains = []
            for domains_item_data in self.domains:
                domains_item = domains_item_data.to_dict()
                domains.append(domains_item)

        services: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.services, Unset):
            services = []
            for services_item_data in self.services:
                services_item = services_item_data.to_dict()
                services.append(services_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if domains is not UNSET:
            field_dict["domains"] = domains
        if services is not UNSET:
            field_dict["services"] = services

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.create_instance_request_domain import CreateInstanceRequestDomain
        from ..models.service import Service

        d = dict(src_dict)
        domains = []
        _domains = d.pop("domains", UNSET)
        for domains_item_data in _domains or []:
            domains_item = CreateInstanceRequestDomain.from_dict(domains_item_data)

            domains.append(domains_item)

        services = []
        _services = d.pop("services", UNSET)
        for services_item_data in _services or []:
            services_item = Service.from_dict(services_item_data)

            services.append(services_item)

        create_instance_request_service_group = cls(
            domains=domains,
            services=services,
        )

        create_instance_request_service_group.additional_properties = d
        return create_instance_request_service_group

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Literal, cast

AutoscalePolicyMetric = Literal["cpu"]

AUTOSCALE_POLICY_METRIC_VALUES: set[AutoscalePolicyMetric] = {
    "cpu",
}


def check_autoscale_policy_metric(value: str) -> AutoscalePolicyMetric:
    if value in AUTOSCALE_POLICY_METRIC_VALUES:
        return cast(AutoscalePolicyMetric, value)
    raise TypeError(f"Unexpected value {value!r}. Expected one of {AUTOSCALE_POLICY_METRIC_VALUES!r}")



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.instance_service_group_instance_domain import InstanceServiceGroupInstanceDomain


T = TypeVar("T", bound="InstanceInstanceServiceGroup")


@_attrs_define
class InstanceInstanceServiceGroup:
    """The service group configuration for the instance.

    This is a reference to the service group that the instance is part of.  The
    service group defines the services (e.g. ports, connection handling) that
    the instance exposes and how they are configured.

    """

    uuid: Union[Unset, UUID] = UNSET
    """ The UUID of the service group.

    This is a unique identifier for the service group that is generated when
    the service is created.  The UUID is used to reference the service group
    in API calls and can be used to identify the service in all API calls
    that require an service identifier. """
    name: Union[Unset, str] = UNSET
    """ The name of the service group.

    This is a human-readable name that can be used to identify the service
    group.  The name is unique within the context of your account.  The name
    can also be used to identify the service group in API calls. """
    domains: Union[Unset, list["InstanceServiceGroupInstanceDomain"]] = UNSET
    """ The domain configuration for the service group. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        uuid: Union[Unset, str] = UNSET
        if not isinstance(self.uuid, Unset):
            uuid = str(self.uuid)

        name = self.name

        domains: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.domains, Unset):
            domains = []
            for domains_item_data in self.domains:
                domains_item = domains_item_data.to_dict()
                domains.append(domains_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if uuid is not UNSET:
            field_dict["uuid"] = uuid
        if name is not UNSET:
            field_dict["name"] = name
        if domains is not UNSET:
            field_dict["domains"] = domains

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.instance_service_group_instance_domain import InstanceServiceGroupInstanceDomain

        d = dict(src_dict)
        _uuid = d.pop("uuid", UNSET)
        uuid: Union[Unset, UUID]
        if isinstance(_uuid, Unset):
            uuid = UNSET
        else:
            uuid = UUID(_uuid)

        name = d.pop("name", UNSET)

        domains = []
        _domains = d.pop("domains", UNSET)
        for domains_item_data in _domains or []:
            domains_item = InstanceServiceGroupInstanceDomain.from_dict(domains_item_data)

            domains.append(domains_item)

        instance_instance_service_group = cls(
            uuid=uuid,
            name=name,
            domains=domains,
        )

        instance_instance_service_group.additional_properties = d
        return instance_instance_service_group

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="Object")


@_attrs_define
class Object:
    """An object is a single component of an image which is external and can be
    uniquely identified by its digest.

    """

    digest: Union[Unset, str] = UNSET
    """ The digest is a string representation including the hashing
    algorithm and the hash value separated by a colon. """
    media_type: Union[Unset, str] = UNSET
    """ The media type of the layer is a string that identifies the type of
    content that the layer contains. """
    size: Union[Unset, int] = UNSET
    """ The size of the layer in bytes. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        digest = self.digest

        media_type = self.media_type

        size = self.size

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if digest is not UNSET:
            field_dict["digest"] = digest
        if media_type is not UNSET:
            field_dict["media_type"] = media_type
        if size is not UNSET:
            field_dict["size"] = size

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        digest = d.pop("digest", UNSET)

        media_type = d.pop("media_type", UNSET)

        size = d.pop("size", UNSET)

        object_ = cls(
            digest=digest,
            media_type=media_type,
            size=size,
        )

        object_.additional_properties = d
        return object_

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.healthz_response_data_services import HealthzResponseDataServices


T = TypeVar("T", bound="HealthzResponseData")


@_attrs_define
class HealthzResponseData:
    """For now, no additional data is returned by the health check."""

    services: Union[Unset, "HealthzResponseDataServices"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        services: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.services, Unset):
            services = self.services.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if services is not UNSET:
            field_dict["services"] = services

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.healthz_response_data_services import HealthzResponseDataServices

        d = dict(src_dict)
        _services = d.pop("services", UNSET)
        services: Union[Unset, HealthzResponseDataServices]
        if isinstance(_services, Unset):
            services = UNSET
        else:
            services = HealthzResponseDataServices.from_dict(_services)

        healthz_response_data = cls(
            services=services,
        )

        healthz_response_data.additional_properties = d
        return healthz_response_data

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, TypeVar, Union
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="InstanceNetworkInterface")


@_attrs_define
class InstanceNetworkInterface:
    """An instance network interface."""

    uuid: Union[Unset, UUID] = UNSET
    """ The UUID of the network interface. This is a unique identifier for the
    network interface that is generated when the instance is created. """
    private_ip: Union[Unset, str] = UNSET
    """ The private IP address of the network interface. This is the internal IP
    address that is used for communication between instances within the same
    network. """
    mac: Union[Unset, str] = UNSET
    """ The MAC address of the network interface. """
    rx_bytes: Union[Unset, int] = UNSET
    """ Amount of bytes received from interface. """
    rx_packets: Union[Unset, int] = UNSET
    """ Count of packets received from interface """
    tx_bytes: Union[Unset, int] = UNSET
    """ Amount of bytes sent to interface. """
    tx_packets: Union[Unset, int] = UNSET
    """ Count of packets sent to interface """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        uuid: Union[Unset, str] = UNSET
        if not isinstance(self.uuid, Unset):
            uuid = str(self.uuid)

        private_ip = self.private_ip

        mac = self.mac

        rx_bytes = self.rx_bytes

        rx_packets = self.rx_packets

        tx_bytes = self.tx_bytes

        tx_packets = self.tx_packets

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if uuid is not UNSET:
            field_dict["uuid"] = uuid
        if private_ip is not UNSET:
            field_dict["private_ip"] = private_ip
        if mac is not UNSET:
            field_dict["mac"] = mac
        if rx_bytes is not UNSET:
            field_dict["rx_bytes"] = rx_bytes
        if rx_packets is not UNSET:
            field_dict["rx_packets"] = rx_packets
        if tx_bytes is not UNSET:
            field_dict["tx_bytes"] = tx_bytes
        if tx_packets is not UNSET:
            field_dict["tx_packets"] = tx_packets

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        _uuid = d.pop("uuid", UNSET)
        uuid: Union[Unset, UUID]
        if isinstance(_uuid, Unset):
            uuid = UNSET
        else:
            uuid = UUID(_uuid)

        private_ip = d.pop("private_ip", UNSET)

        mac = d.pop("mac", UNSET)

        rx_bytes = d.pop("rx_bytes", UNSET)

        rx_packets = d.pop("rx_packets", UNSET)

        tx_bytes = d.pop("tx_bytes", UNSET)

        tx_packets = d.pop("tx_packets", UNSET)

        instance_network_interface = cls(
            uuid=uuid,
            private_ip=private_ip,
            mac=mac,
            rx_bytes=rx_bytes,
            rx_packets=rx_packets,
            tx_bytes=tx_bytes,
            tx_packets=tx_packets,
        )

        instance_network_interface.additional_properties = d
        return instance_network_interface

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.response_status import ResponseStatus, check_response_status
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.domain import Domain
    from ..models.service import Service
    from ..models.service_group_instance import ServiceGroupInstance


T = TypeVar("T", bound="ServiceGroup")


@_attrs_define
class ServiceGroup:
    """A service group on Unikraft Cloud is used to describe how your application
    exposes its functionality to the outside world.  Once defined, assigning an
    instance to the service will make it accessible from the Internet.

    An application, running as an instance, may expose one or more ports, e.g. it
    listens on port 80 because your application exposes a HTTP web service. This,
    along with a set of additional metadata defines how the "service" is
    configured and accessed.  For example, a service may be configured to use
    TLS, or be bound to a specific domain name.

    When an instance is assigned to a service group, it immediately becomes
    accessible over the Internet on the exposed public port, using the set DNS
    name, and is routed to the set destination port.

    Note: If you do not specify a DNS name when you create a service and you
    indicate that the application exposes some ports, Unikraft Cloud will
    generates a random DNS name for you.  Unikraft Cloud also supports custom
    domains like www.example.com and wildcard domains like *.example.com.

    """

    uuid: Union[Unset, UUID] = UNSET
    """ The UUID of the service group.

    This is a unique identifier for the service group that is generated when
    the service group is created.  The UUID is used to reference the service in
    API calls and can be used to identify the service group in all API calls
    that require an identifier. """
    name: Union[Unset, str] = UNSET
    """ The name of the service group.

    This is a human-readable name that can be used to identify the service
    group. The name must be unique within the context of your account.  The
    name can also be used to identify the service in API calls. """
    created_at: Union[Unset, datetime.datetime] = UNSET
    """ The time the service was created. """
    persistent: Union[Unset, bool] = UNSET
    """ Indicates if the service will stay remain even after the last instance
    detached.  If this is set to false, the service will be deleted when the
    last instance detached from it.  If this is set to true, the service will
    remain and can be reused by other instances.  This is useful if you want to
    keep the service configuration, e.g., the published ports, handlers, and
    domains, even if there are no instances assigned to it. """
    autoscale: Union[Unset, bool] = UNSET
    """ Indicates if the service has autoscale enabled.  See the associated
    autoscale documentation for more information about how to set this up.
    Autoscale policies can be set up after the service has been created. """
    soft_limit: Union[Unset, int] = UNSET
    """ The soft limit is used by the Unikraft Cloud load balancer to decide when
    to wake up another standby instance.  For example, if the soft limit is set
    to 5 and the service consists of 2 standby instances, one of the instances
    receives up to 5 concurrent requests.  The 6th parallel requests wakes up
    the second instance.  If there are no more standby instances to wake up,
    the number of requests assigned to each instance will exceed the soft
    limit.  The load balancer makes sure that when the number of in-flight
    requests goes down again, instances are put into standby as fast as
    possible. """
    hard_limit: Union[Unset, int] = UNSET
    """ The hard limit defines the maximum number of concurrent requests that an
    instance assigned to the this service can handle.  The load balancer will
    never assign more requests to a single instance.  In case there are no
    other instances available, excess requests fail (i.e., they are blocked and
    not queued). """
    services: Union[Unset, list["Service"]] = UNSET
    """ List of published network ports for this service and the destination port
    to which Unikraft Cloud will forward traffic to.  Additional handlers can
    be defined for each published port in order to define how the service will
    handle incoming connections and forward traffic from the Internet to your
    application.  For example, a service can be configured to terminate TLS
    connections, redirect HTTP traffic, or enable HTTP mode for load balancing. """
    domains: Union[Unset, list["Domain"]] = UNSET
    """ List of domains associated with the service.  Domains are used to access
    the service over the Internet. """
    instances: Union[Unset, list["ServiceGroupInstance"]] = UNSET
    """ List of instances assigned to the service. """
    status: Union[Unset, ResponseStatus] = UNSET
    """ The response status of an API request. """
    message: Union[Unset, str] = UNSET
    """ An optional message providing additional information about the status.
    This field is only set when this message object is used as a response
    message, and is useful when the status is not `success`. """
    error: Union[Unset, int] = UNSET
    """ An optional error code providing additional information about the status.
    This field is only set when this message object is used as a response
    message, and is useful when the status is not `success`. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        uuid: Union[Unset, str] = UNSET
        if not isinstance(self.uuid, Unset):
            uuid = str(self.uuid)

        name = self.name

        created_at: Union[Unset, str] = UNSET
        if not isinstance(self.created_at, Unset):
            created_at = self.created_at.isoformat()

        persistent = self.persistent

        autoscale = self.autoscale

        soft_limit = self.soft_limit

        hard_limit = self.hard_limit

        services: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.services, Unset):
            services = []
            for services_item_data in self.services:
                services_item = services_item_data.to_dict()
                services.append(services_item)

        domains: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.domains, Unset):
            domains = []
            for domains_item_data in self.domains:
                domains_item = domains_item_data.to_dict()
                domains.append(domains_item)

        instances: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.instances, Unset):
            instances = []
            for instances_item_data in self.instances:
                instances_item = instances_item_data.to_dict()
                instances.append(instances_item)

        status: Union[Unset, str] = UNSET
        if not isinstance(self.status, Unset):
            status = self.status

        message = self.message

        error = self.error

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if uuid is not UNSET:
            field_dict["uuid"] = uuid
        if name is not UNSET:
            field_dict["name"] = name
        if created_at is not UNSET:
            field_dict["created_at"] = created_at
        if persistent is not UNSET:
            field_dict["persistent"] = persistent
        if autoscale is not UNSET:
            field_dict["autoscale"] = autoscale
        if soft_limit is not UNSET:
            field_dict["soft_limit"] = soft_limit
        if hard_limit is not UNSET:
            field_dict["hard_limit"] = hard_limit
        if services is not UNSET:
            field_dict["services"] = services
        if domains is not UNSET:
            field_dict["domains"] = domains
        if instances is not UNSET:
            field_dict["instances"] = instances
        if status is not UNSET:
            field_dict["status"] = status
        if message is not UNSET:
            field_dict["message"] = message
        if error is not UNSET:
            field_dict["error"] = error

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.domain import Domain
        from ..models.service import Service
        from ..models.service_group_instance import ServiceGroupInstance

        d = dict(src_dict)
        _uuid = d.pop("uuid", UNSET)
        uuid: Union[Unset, UUID]
        if isinstance(_uuid, Unset):
            uuid = UNSET
        else:
            uuid = UUID(_uuid)

        name = d.pop("name", UNSET)

        _created_at = d.pop("created_at", UNSET)
        created_at: Union[Unset, datetime.datetime]
        if isinstance(_created_at, Unset):
            created_at = UNSET
        else:
            created_at = isoparse(_created_at)

        persistent = d.pop("persistent", UNSET)

        autoscale = d.pop("autoscale", UNSET)

        soft_limit = d.pop("soft_limit", UNSET)

        hard_limit = d.pop("hard_limit", UNSET)

        services = []
        _services = d.pop("services", UNSET)
        for services_item_data in _services or []:
            services_item = Service.from_dict(services_item_data)

            services.append(services_item)

        domains = []
        _domains = d.pop("domains", UNSET)
        for domains_item_data in _domains or []:
            domains_item = Domain.from_dict(domains_item_data)

            domains.append(domains_item)

        instances = []
        _instances = d.pop("instances", UNSET)
        for instances_item_data in _instances or []:
            instances_item = ServiceGroupInstance.from_dict(instances_item_data)

            instances.append(instances_item)

        _status = d.pop("status", UNSET)
        status: Union[Unset, ResponseStatus]
        if isinstance(_status, Unset):
            status = UNSET
        else:
            status = check_response_status(_status)

        message = d.pop("message", UNSET)

        error = d.pop("error", UNSET)

        service_group = cls(
            uuid=uuid,
            name=name,
            created_at=created_at,
            persistent=persistent,
            autoscale=autoscale,
            soft_limit=soft_limit,
            hard_limit=hard_limit,
            services=services,
            domains=domains,
            instances=instances,
            status=status,
            message=message,
            error=error,
        )

        service_group.additional_properties = d
        return service_group

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Literal, cast

ServiceHandlersItem = Literal["http", "redirect", "tls"]

SERVICE_HANDLERS_ITEM_VALUES: set[ServiceHandlersItem] = {
    "http",
    "redirect",
    "tls",
}


def check_service_handlers_item(value: str) -> ServiceHandlersItem:
    if value in SERVICE_HANDLERS_ITEM_VALUES:
        return cast(ServiceHandlersItem, value)
    raise TypeError(f"Unexpected value {value!r}. Expected one of {SERVICE_HANDLERS_ITEM_VALUES!r}")



from typing import Literal, cast

UpdateServiceGroupsRequestItemOp = Literal["add", "del", "set"]

UPDATE_SERVICE_GROUPS_REQUEST_ITEM_OP_VALUES: set[UpdateServiceGroupsRequestItemOp] = {
    "add",
    "del",
    "set",
}


def check_update_service_groups_request_item_op(value: str) -> UpdateServiceGroupsRequestItemOp:
    if value in UPDATE_SERVICE_GROUPS_REQUEST_ITEM_OP_VALUES:
        return cast(UpdateServiceGroupsRequestItemOp, value)
    raise TypeError(f"Unexpected value {value!r}. Expected one of {UPDATE_SERVICE_GROUPS_REQUEST_ITEM_OP_VALUES!r}")



import datetime
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.response_status import ResponseStatus, check_response_status
from ..models.volume_state import VolumeState, check_volume_state
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.volume_instance_id import VolumeInstanceID
    from ..models.volume_volume_instance_mount import VolumeVolumeInstanceMount


T = TypeVar("T", bound="Volume")


@_attrs_define
class Volume:
    """A volume represents a storage device that can be attached to an instance."""

    uuid: Union[Unset, UUID] = UNSET
    """ The UUID of the volume.

    This is a unique identifier for the volume that is generated when the
    volume is created.  The UUID is used to reference the volume in
    API calls and can be used to identify the volume in all API calls that
    require an identifier. """
    name: Union[Unset, str] = UNSET
    """ The name of the volume.

    This is a human-readable name that can be used to identify the volume.
    The name must be unique within the context of your account.  The name can
    also be used to identify the volume in API calls. """
    created_at: Union[Unset, datetime.datetime] = UNSET
    """ The time the volume was created. """
    state: Union[Unset, VolumeState] = UNSET
    """ Current state of the volume. """
    size_mb: Union[Unset, int] = UNSET
    """ The size of the volume in megabytes. """
    persistent: Union[Unset, bool] = UNSET
    """ Indicates if the volume will stay alive when the last instance is deleted
    that this volume is attached to. """
    attached_to: Union[Unset, list["VolumeInstanceID"]] = UNSET
    """ List of instances that this volume is attached to. """
    mounted_by: Union[Unset, list["VolumeVolumeInstanceMount"]] = UNSET
    """ List of instances that have this volume mounted. """
    tags: Union[Unset, list[str]] = UNSET
    """ The tags associated with the volume. """
    status: Union[Unset, ResponseStatus] = UNSET
    """ The response status of an API request. """
    message: Union[Unset, str] = UNSET
    """ An optional message providing additional information about the status.
    This field is only set when this message object is used as a response
    message, and is useful when the status is not `success`. """
    error: Union[Unset, int] = UNSET
    """ An optional error code providing additional information about the status.
    This field is only set when this message object is used as a response
    message, and is useful when the status is not `success`. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        uuid: Union[Unset, str] = UNSET
        if not isinstance(self.uuid, Unset):
            uuid = str(self.uuid)

        name = self.name

        created_at: Union[Unset, str] = UNSET
        if not isinstance(self.created_at, Unset):
            created_at = self.created_at.isoformat()

        state: Union[Unset, str] = UNSET
        if not isinstance(self.state, Unset):
            state = self.state

        size_mb = self.size_mb

        persistent = self.persistent

        attached_to: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.attached_to, Unset):
            attached_to = []
            for attached_to_item_data in self.attached_to:
                attached_to_item = attached_to_item_data.to_dict()
                attached_to.append(attached_to_item)

        mounted_by: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.mounted_by, Unset):
            mounted_by = []
            for mounted_by_item_data in self.mounted_by:
                mounted_by_item = mounted_by_item_data.to_dict()
                mounted_by.append(mounted_by_item)

        tags: Union[Unset, list[str]] = UNSET
        if not isinstance(self.tags, Unset):
            tags = self.tags

        status: Union[Unset, str] = UNSET
        if not isinstance(self.status, Unset):
            status = self.status

        message = self.message

        error = self.error

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if uuid is not UNSET:
            field_dict["uuid"] = uuid
        if name is not UNSET:
            field_dict["name"] = name
        if created_at is not UNSET:
            field_dict["created_at"] = created_at
        if state is not UNSET:
            field_dict["state"] = state
        if size_mb is not UNSET:
            field_dict["size_mb"] = size_mb
        if persistent is not UNSET:
            field_dict["persistent"] = persistent
        if attached_to is not UNSET:
            field_dict["attached_to"] = attached_to
        if mounted_by is not UNSET:
            field_dict["mounted_by"] = mounted_by
        if tags is not UNSET:
            field_dict["tags"] = tags
        if status is not UNSET:
            field_dict["status"] = status
        if message is not UNSET:
            field_dict["message"] = message
        if error is not UNSET:
            field_dict["error"] = error

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.volume_instance_id import VolumeInstanceID
        from ..models.volume_volume_instance_mount import VolumeVolumeInstanceMount

        d = dict(src_dict)
        _uuid = d.pop("uuid", UNSET)
        uuid: Union[Unset, UUID]
        if isinstance(_uuid, Unset):
            uuid = UNSET
        else:
            uuid = UUID(_uuid)

        name = d.pop("name", UNSET)

        _created_at = d.pop("created_at", UNSET)
        created_at: Union[Unset, datetime.datetime]
        if isinstance(_created_at, Unset):
            created_at = UNSET
        else:
            created_at = isoparse(_created_at)

        _state = d.pop("state", UNSET)
        state: Union[Unset, VolumeState]
        if isinstance(_state, Unset):
            state = UNSET
        else:
            state = check_volume_state(_state)

        size_mb = d.pop("size_mb", UNSET)

        persistent = d.pop("persistent", UNSET)

        attached_to = []
        _attached_to = d.pop("attached_to", UNSET)
        for attached_to_item_data in _attached_to or []:
            attached_to_item = VolumeInstanceID.from_dict(attached_to_item_data)

            attached_to.append(attached_to_item)

        mounted_by = []
        _mounted_by = d.pop("mounted_by", UNSET)
        for mounted_by_item_data in _mounted_by or []:
            mounted_by_item = VolumeVolumeInstanceMount.from_dict(mounted_by_item_data)

            mounted_by.append(mounted_by_item)

        tags = cast(list[str], d.pop("tags", UNSET))

        _status = d.pop("status", UNSET)
        status: Union[Unset, ResponseStatus]
        if isinstance(_status, Unset):
            status = UNSET
        else:
            status = check_response_status(_status)

        message = d.pop("message", UNSET)

        error = d.pop("error", UNSET)

        volume = cls(
            uuid=uuid,
            name=name,
            created_at=created_at,
            state=state,
            size_mb=size_mb,
            persistent=persistent,
            attached_to=attached_to,
            mounted_by=mounted_by,
            tags=tags,
            status=status,
            message=message,
            error=error,
        )

        volume.additional_properties = d
        return volume

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="GetInstanceMetricsResponseInstanceMetrics")


@_attrs_define
class GetInstanceMetricsResponseInstanceMetrics:
    rss_bytes: Union[Unset, int] = UNSET
    """ Resident set size of the VMM in bytes.

    The resident set size (RSS) specifies the amount of physical memory that
    has been touched by the instance and is currently reserved for the
    instance on the Unikraft Cloud server.  The RSS grows until the instance
    has touched all memory assigned to it via the memory_mb setting and may
    also exceed this value as supporting services running outside the
    instance acquire memory.  The RSS is different from the current amount of
    memory allocated by the application, which is likely to fluctuate over
    the lifetime of the application.  The RSS is not a cumulative metric.
    When the instance is stopped rss goes down to 0. """
    cpu_time_ms: Union[Unset, int] = UNSET
    """ Consumed CPU time in milliseconds. """
    boot_time_us: Union[Unset, int] = UNSET
    """ The boot time of the instance in microseconds.  We take a pragmatic
    approach is to define the boot time.  We calculate this as the difference
    in time between the moment the virtualization toolstack is invoked to
    respond to a VM boot request and the moment the OS starts executing user
    code (i.e., the end of the guest OS boot process).  This is essentially the
    time that a user would experience in a deployment, minus the application
    initialization time, which we leave out since it is independent from the
    OS. """
    net_time_us: Union[Unset, int] = UNSET
    """ This is the time it took for the user-level application to start listening
    on a non-localhost port measured in microseconds.  This is the time from
    when the instance started until it reasonably ready to start responding to
    network requests.  This is useful for measuring the time it takes for the
    instance to become operationally ready. """
    rx_bytes: Union[Unset, int] = UNSET
    """ Total amount of bytes received from network. """
    rx_packets: Union[Unset, int] = UNSET
    """ Total count of packets received from network. """
    tx_bytes: Union[Unset, int] = UNSET
    """ Total amount of bytes transmitted over network. """
    tx_packets: Union[Unset, int] = UNSET
    """ Total count of packets transmitted over network. """
    nconns: Union[Unset, int] = UNSET
    """ Number of currently established inbound connections (non-HTTP). """
    nreqs: Union[Unset, int] = UNSET
    """ Number of in-flight HTTP requests. """
    nqueued: Union[Unset, int] = UNSET
    """ Number of queued inbound connections and HTTP requests. """
    ntotal: Union[Unset, int] = UNSET
    """ Total number of inbound connections and HTTP requests handled. """
    message: Union[Unset, str] = UNSET
    """ An optional message providing additional information about the status.
    This field is useful when the status is not `success`. """
    error: Union[Unset, int] = UNSET
    """ An optional error code providing additional information about the status.
    This field is useful when the status is not `success`. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        rss_bytes = self.rss_bytes

        cpu_time_ms = self.cpu_time_ms

        boot_time_us = self.boot_time_us

        net_time_us = self.net_time_us

        rx_bytes = self.rx_bytes

        rx_packets = self.rx_packets

        tx_bytes = self.tx_bytes

        tx_packets = self.tx_packets

        nconns = self.nconns

        nreqs = self.nreqs

        nqueued = self.nqueued

        ntotal = self.ntotal

        message = self.message

        error = self.error

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if rss_bytes is not UNSET:
            field_dict["rss_bytes"] = rss_bytes
        if cpu_time_ms is not UNSET:
            field_dict["cpu_time_ms"] = cpu_time_ms
        if boot_time_us is not UNSET:
            field_dict["boot_time_us"] = boot_time_us
        if net_time_us is not UNSET:
            field_dict["net_time_us"] = net_time_us
        if rx_bytes is not UNSET:
            field_dict["rx_bytes"] = rx_bytes
        if rx_packets is not UNSET:
            field_dict["rx_packets"] = rx_packets
        if tx_bytes is not UNSET:
            field_dict["tx_bytes"] = tx_bytes
        if tx_packets is not UNSET:
            field_dict["tx_packets"] = tx_packets
        if nconns is not UNSET:
            field_dict["nconns"] = nconns
        if nreqs is not UNSET:
            field_dict["nreqs"] = nreqs
        if nqueued is not UNSET:
            field_dict["nqueued"] = nqueued
        if ntotal is not UNSET:
            field_dict["ntotal"] = ntotal
        if message is not UNSET:
            field_dict["message"] = message
        if error is not UNSET:
            field_dict["error"] = error

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        rss_bytes = d.pop("rss_bytes", UNSET)

        cpu_time_ms = d.pop("cpu_time_ms", UNSET)

        boot_time_us = d.pop("boot_time_us", UNSET)

        net_time_us = d.pop("net_time_us", UNSET)

        rx_bytes = d.pop("rx_bytes", UNSET)

        rx_packets = d.pop("rx_packets", UNSET)

        tx_bytes = d.pop("tx_bytes", UNSET)

        tx_packets = d.pop("tx_packets", UNSET)

        nconns = d.pop("nconns", UNSET)

        nreqs = d.pop("nreqs", UNSET)

        nqueued = d.pop("nqueued", UNSET)

        ntotal = d.pop("ntotal", UNSET)

        message = d.pop("message", UNSET)

        error = d.pop("error", UNSET)

        get_instance_metrics_response_instance_metrics = cls(
            rss_bytes=rss_bytes,
            cpu_time_ms=cpu_time_ms,
            boot_time_us=boot_time_us,
            net_time_us=net_time_us,
            rx_bytes=rx_bytes,
            rx_packets=rx_packets,
            tx_bytes=tx_bytes,
            tx_packets=tx_packets,
            nconns=nconns,
            nreqs=nreqs,
            nqueued=nqueued,
            ntotal=ntotal,
            message=message,
            error=error,
        )

        get_instance_metrics_response_instance_metrics.additional_properties = d
        return get_instance_metrics_response_instance_metrics

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, TypeVar, Union
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="BodyInstanceID")


@_attrs_define
class BodyInstanceID:
    """Reference to the instance to detach the volume from."""

    uuid: Union[Unset, UUID] = UNSET
    """ The UUID of the instance that the volume is detached from. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        uuid: Union[Unset, str] = UNSET
        if not isinstance(self.uuid, Unset):
            uuid = str(self.uuid)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if uuid is not UNSET:
            field_dict["uuid"] = uuid

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        _uuid = d.pop("uuid", UNSET)
        uuid: Union[Unset, UUID]
        if isinstance(_uuid, Unset):
            uuid = UNSET
        else:
            uuid = UUID(_uuid)

        body_instance_id = cls(
            uuid=uuid,
        )

        body_instance_id.additional_properties = d
        return body_instance_id

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="QuotasLimits")


@_attrs_define
class QuotasLimits:
    min_memory_mb: Union[Unset, int] = UNSET
    """ Minimum amount of memory assigned to live instances in megabytes """
    max_memory_mb: Union[Unset, int] = UNSET
    """ Maximum amount of memory assigned to live instances in megabytes """
    min_volume_mb: Union[Unset, int] = UNSET
    """ Minimum size of a volume in megabytes """
    max_volume_mb: Union[Unset, int] = UNSET
    """ Maximum size of a volume in megabytes """
    min_autoscale_size: Union[Unset, int] = UNSET
    """ Minimum size of an autoscale group """
    max_autoscale_size: Union[Unset, int] = UNSET
    """ Maximum size of an autoscale group """
    min_vcpus: Union[Unset, int] = UNSET
    """ Minimum number of vCPUs """
    max_vcpus: Union[Unset, int] = UNSET
    """ Maximum number of vCPUs """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        min_memory_mb = self.min_memory_mb

        max_memory_mb = self.max_memory_mb

        min_volume_mb = self.min_volume_mb

        max_volume_mb = self.max_volume_mb

        min_autoscale_size = self.min_autoscale_size

        max_autoscale_size = self.max_autoscale_size

        min_vcpus = self.min_vcpus

        max_vcpus = self.max_vcpus

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if min_memory_mb is not UNSET:
            field_dict["min_memory_mb"] = min_memory_mb
        if max_memory_mb is not UNSET:
            field_dict["max_memory_mb"] = max_memory_mb
        if min_volume_mb is not UNSET:
            field_dict["min_volume_mb"] = min_volume_mb
        if max_volume_mb is not UNSET:
            field_dict["max_volume_mb"] = max_volume_mb
        if min_autoscale_size is not UNSET:
            field_dict["min_autoscale_size"] = min_autoscale_size
        if max_autoscale_size is not UNSET:
            field_dict["max_autoscale_size"] = max_autoscale_size
        if min_vcpus is not UNSET:
            field_dict["min_vcpus"] = min_vcpus
        if max_vcpus is not UNSET:
            field_dict["max_vcpus"] = max_vcpus

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        min_memory_mb = d.pop("min_memory_mb", UNSET)

        max_memory_mb = d.pop("max_memory_mb", UNSET)

        min_volume_mb = d.pop("min_volume_mb", UNSET)

        max_volume_mb = d.pop("max_volume_mb", UNSET)

        min_autoscale_size = d.pop("min_autoscale_size", UNSET)

        max_autoscale_size = d.pop("max_autoscale_size", UNSET)

        min_vcpus = d.pop("min_vcpus", UNSET)

        max_vcpus = d.pop("max_vcpus", UNSET)

        quotas_limits = cls(
            min_memory_mb=min_memory_mb,
            max_memory_mb=max_memory_mb,
            min_volume_mb=min_volume_mb,
            max_volume_mb=max_volume_mb,
            min_autoscale_size=min_autoscale_size,
            max_autoscale_size=max_autoscale_size,
            min_vcpus=min_vcpus,
            max_vcpus=max_vcpus,
        )

        quotas_limits.additional_properties = d
        return quotas_limits

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Optional, Union
from uuid import UUID

import httpx

from ...client import AuthenticatedClient, Client
from ...models.delete_instances_response import DeleteInstancesResponse
from ...types import Response


def _get_kwargs(
    uuid: UUID,
) -> dict[str, Any]:
    _kwargs: dict[str, Any] = {
        "method": "delete",
        "url": f"/v1/instances/{uuid}",
    }

    return _kwargs


def _parse_response(*, client: Union[AuthenticatedClient, Client], response: httpx.Response) -> DeleteInstancesResponse:
    response_default = DeleteInstancesResponse.from_dict(response.json())

    return response_default


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[DeleteInstancesResponse]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    uuid: UUID,
    *,
    client: Union[AuthenticatedClient, Client],
) -> Response[DeleteInstancesResponse]:
    """Delete Instance by UUID

     Delete a specified instance by its UUID.  After this call the UUID of the
    instance is no longer valid.  If the instance is currently running,
    it is force-stopped.

    Args:
        uuid (UUID):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[DeleteInstancesResponse]
    """

    kwargs = _get_kwargs(
        uuid=uuid,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    uuid: UUID,
    *,
    client: Union[AuthenticatedClient, Client],
) -> Optional[DeleteInstancesResponse]:
    """Delete Instance by UUID

     Delete a specified instance by its UUID.  After this call the UUID of the
    instance is no longer valid.  If the instance is currently running,
    it is force-stopped.

    Args:
        uuid (UUID):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        DeleteInstancesResponse
    """

    return sync_detailed(
        uuid=uuid,
        client=client,
    ).parsed


async def asyncio_detailed(
    uuid: UUID,
    *,
    client: Union[AuthenticatedClient, Client],
) -> Response[DeleteInstancesResponse]:
    """Delete Instance by UUID

     Delete a specified instance by its UUID.  After this call the UUID of the
    instance is no longer valid.  If the instance is currently running,
    it is force-stopped.

    Args:
        uuid (UUID):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[DeleteInstancesResponse]
    """

    kwargs = _get_kwargs(
        uuid=uuid,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    uuid: UUID,
    *,
    client: Union[AuthenticatedClient, Client],
) -> Optional[DeleteInstancesResponse]:
    """Delete Instance by UUID

     Delete a specified instance by its UUID.  After this call the UUID of the
    instance is no longer valid.  If the instance is currently running,
    it is force-stopped.

    Args:
        uuid (UUID):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        DeleteInstancesResponse
    """

    return (
        await asyncio_detailed(
            uuid=uuid,
            client=client,
        )
    ).parsed



from collections.abc import Mapping
from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="QuotasStats")


@_attrs_define
class QuotasStats:
    instances: Union[Unset, int] = UNSET
    """ Number of instances """
    live_instances: Union[Unset, int] = UNSET
    """ Number of instances that are not in the `stopped` state """
    live_vcpus: Union[Unset, int] = UNSET
    """ Number of vCPUs """
    live_memory_mb: Union[Unset, int] = UNSET
    """ Amount of memory assigned to instances that are not in the `stopped`
    state in megabytes """
    service_groups: Union[Unset, int] = UNSET
    """ Number of services """
    services: Union[Unset, int] = UNSET
    """ Number of published network ports over all existing services """
    volumes: Union[Unset, int] = UNSET
    """ Number of volumes """
    total_volume_mb: Union[Unset, int] = UNSET
    """ Total size of all volumes in megabytes """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        instances = self.instances

        live_instances = self.live_instances

        live_vcpus = self.live_vcpus

        live_memory_mb = self.live_memory_mb

        service_groups = self.service_groups

        services = self.services

        volumes = self.volumes

        total_volume_mb = self.total_volume_mb

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if instances is not UNSET:
            field_dict["instances"] = instances
        if live_instances is not UNSET:
            field_dict["live_instances"] = live_instances
        if live_vcpus is not UNSET:
            field_dict["live_vcpus"] = live_vcpus
        if live_memory_mb is not UNSET:
            field_dict["live_memory_mb"] = live_memory_mb
        if service_groups is not UNSET:
            field_dict["service_groups"] = service_groups
        if services is not UNSET:
            field_dict["services"] = services
        if volumes is not UNSET:
            field_dict["volumes"] = volumes
        if total_volume_mb is not UNSET:
            field_dict["total_volume_mb"] = total_volume_mb

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        instances = d.pop("instances", UNSET)

        live_instances = d.pop("live_instances", UNSET)

        live_vcpus = d.pop("live_vcpus", UNSET)

        live_memory_mb = d.pop("live_memory_mb", UNSET)

        service_groups = d.pop("service_groups", UNSET)

        services = d.pop("services", UNSET)

        volumes = d.pop("volumes", UNSET)

        total_volume_mb = d.pop("total_volume_mb", UNSET)

        quotas_stats = cls(
            instances=instances,
            live_instances=live_instances,
            live_vcpus=live_vcpus,
            live_memory_mb=live_memory_mb,
            service_groups=service_groups,
            services=services,
            volumes=volumes,
            total_volume_mb=total_volume_mb,
        )

        quotas_stats.additional_properties = d
        return quotas_stats

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Literal, cast

InstanceRestartPolicy = Literal["always", "never", "on_failure"]

INSTANCE_RESTART_POLICY_VALUES: set[InstanceRestartPolicy] = {
    "always",
    "never",
    "on_failure",
}


def check_instance_restart_policy(value: str) -> InstanceRestartPolicy:
    if value in INSTANCE_RESTART_POLICY_VALUES:
        return cast(InstanceRestartPolicy, value)
    raise TypeError(f"Unexpected value {value!r}. Expected one of {INSTANCE_RESTART_POLICY_VALUES!r}")



from typing import Literal, cast

WaitInstanceResponseWaitedInstanceState = Literal["draining", "running", "standby", "starting", "stopped", "stopping"]

WAIT_INSTANCE_RESPONSE_WAITED_INSTANCE_STATE_VALUES: set[WaitInstanceResponseWaitedInstanceState] = {
    "draining",
    "running",
    "standby",
    "starting",
    "stopped",
    "stopping",
}


def check_wait_instance_response_waited_instance_state(value: str) -> WaitInstanceResponseWaitedInstanceState:
    if value in WAIT_INSTANCE_RESPONSE_WAITED_INSTANCE_STATE_VALUES:
        return cast(WaitInstanceResponseWaitedInstanceState, value)
    raise TypeError(
        f"Unexpected value {value!r}. Expected one of {WAIT_INSTANCE_RESPONSE_WAITED_INSTANCE_STATE_VALUES!r}"
    )



from collections.abc import Mapping
from typing import Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="CreateCertificateRequest")


@_attrs_define
class CreateCertificateRequest:
    """The request message for creating/uploading a new certificate."""

    cn: str
    """ The common name (CN) of the certificate. """
    chain: str
    """ The chain of the certificate. """
    pkey: str
    """ The private key of the certificate. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        cn = self.cn

        chain = self.chain

        pkey = self.pkey

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "cn": cn,
                "chain": chain,
                "pkey": pkey,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        cn = d.pop("cn")

        chain = d.pop("chain")

        pkey = d.pop("pkey")

        create_certificate_request = cls(
            cn=cn,
            chain=chain,
            pkey=pkey,
        )

        create_certificate_request.additional_properties = d
        return create_certificate_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.response_status import ResponseStatus, check_response_status
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.autoscale_policy import AutoscalePolicy


T = TypeVar("T", bound="GetAutoscaleConfigurationPolicyResponsePolicyResponse")


@_attrs_define
class GetAutoscaleConfigurationPolicyResponsePolicyResponse:
    status: Union[Unset, ResponseStatus] = UNSET
    """ The response status of an API request. """
    policy: Union[Unset, "AutoscalePolicy"] = UNSET
    """ AutoscalePolicy defines the autoscale policy for a service.
    Right now it contains fields from both the `ondemand` and `step` policies.
    They are marked both as optional, so only one of them should be set at a
    time. This is a current limitation of the API design. """
    message: Union[Unset, str] = UNSET
    """ An optional message providing additional information about the status.
    This field is useful when the status is not `success`. """
    error: Union[Unset, int] = UNSET
    """ An optional error code providing additional information about the status.
    This field is useful when the status is not `success`. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        status: Union[Unset, str] = UNSET
        if not isinstance(self.status, Unset):
            status = self.status

        policy: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.policy, Unset):
            policy = self.policy.to_dict()

        message = self.message

        error = self.error

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if status is not UNSET:
            field_dict["status"] = status
        if policy is not UNSET:
            field_dict["policy"] = policy
        if message is not UNSET:
            field_dict["message"] = message
        if error is not UNSET:
            field_dict["error"] = error

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.autoscale_policy import AutoscalePolicy

        d = dict(src_dict)
        _status = d.pop("status", UNSET)
        status: Union[Unset, ResponseStatus]
        if isinstance(_status, Unset):
            status = UNSET
        else:
            status = check_response_status(_status)

        _policy = d.pop("policy", UNSET)
        policy: Union[Unset, AutoscalePolicy]
        if isinstance(_policy, Unset):
            policy = UNSET
        else:
            policy = AutoscalePolicy.from_dict(_policy)

        message = d.pop("message", UNSET)

        error = d.pop("error", UNSET)

        get_autoscale_configuration_policy_response_policy_response = cls(
            status=status,
            policy=policy,
            message=message,
            error=error,
        )

        get_autoscale_configuration_policy_response_policy_response.additional_properties = d
        return get_autoscale_configuration_policy_response_policy_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Literal, cast

VolumeState = Literal["available", "busy", "error", "idle", "initializing", "mounted", "uninitialized"]

VOLUME_STATE_VALUES: set[VolumeState] = {
    "available",
    "busy",
    "error",
    "idle",
    "initializing",
    "mounted",
    "uninitialized",
}


def check_volume_state(value: str) -> VolumeState:
    if value in VOLUME_STATE_VALUES:
        return cast(VolumeState, value)
    raise TypeError(f"Unexpected value {value!r}. Expected one of {VOLUME_STATE_VALUES!r}")



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.detach_volumes_request_instance_id import DetachVolumesRequestInstanceID


T = TypeVar("T", bound="DetachVolumesRequest")


@_attrs_define
class DetachVolumesRequest:
    """The request message for detaching one or more volume(s) from instances by
    their UUID(s) or name(s).

    """

    uuid: Union[Unset, UUID] = UNSET
    """ The UUID of the volume to detach. Mutually exclusive with name.
    Exactly one of uuid or name must be provided. """
    name: Union[Unset, str] = UNSET
    """ The name of the volume to detach. Mutually exclusive with UUID.
    Exactly one of uuid or name must be provided. """
    from_: Union[Unset, "DetachVolumesRequestInstanceID"] = UNSET
    """ Reference to the instance to detach the volume from. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        uuid: Union[Unset, str] = UNSET
        if not isinstance(self.uuid, Unset):
            uuid = str(self.uuid)

        name = self.name

        from_: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.from_, Unset):
            from_ = self.from_.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if uuid is not UNSET:
            field_dict["uuid"] = uuid
        if name is not UNSET:
            field_dict["name"] = name
        if from_ is not UNSET:
            field_dict["from"] = from_

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.detach_volumes_request_instance_id import DetachVolumesRequestInstanceID

        d = dict(src_dict)
        _uuid = d.pop("uuid", UNSET)
        uuid: Union[Unset, UUID]
        if isinstance(_uuid, Unset):
            uuid = UNSET
        else:
            uuid = UUID(_uuid)

        name = d.pop("name", UNSET)

        _from_ = d.pop("from", UNSET)
        from_: Union[Unset, DetachVolumesRequestInstanceID]
        if isinstance(_from_, Unset):
            from_ = UNSET
        else:
            from_ = DetachVolumesRequestInstanceID.from_dict(_from_)

        detach_volumes_request = cls(
            uuid=uuid,
            name=name,
            from_=from_,
        )

        detach_volumes_request.additional_properties = d
        return detach_volumes_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Literal, cast

WaitInstanceByUUIDRequestBodyState = Literal["draining", "running", "standby", "starting", "stopped", "stopping"]

WAIT_INSTANCE_BY_UUID_REQUEST_BODY_STATE_VALUES: set[WaitInstanceByUUIDRequestBodyState] = {
    "draining",
    "running",
    "standby",
    "starting",
    "stopped",
    "stopping",
}


def check_wait_instance_by_uuid_request_body_state(value: str) -> WaitInstanceByUUIDRequestBodyState:
    if value in WAIT_INSTANCE_BY_UUID_REQUEST_BODY_STATE_VALUES:
        return cast(WaitInstanceByUUIDRequestBodyState, value)
    raise TypeError(f"Unexpected value {value!r}. Expected one of {WAIT_INSTANCE_BY_UUID_REQUEST_BODY_STATE_VALUES!r}")



from typing import Literal, cast

UpdateServiceGroupByUUIDRequestBodyOp = Literal["add", "del", "set"]

UPDATE_SERVICE_GROUP_BY_UUID_REQUEST_BODY_OP_VALUES: set[UpdateServiceGroupByUUIDRequestBodyOp] = {
    "add",
    "del",
    "set",
}


def check_update_service_group_by_uuid_request_body_op(value: str) -> UpdateServiceGroupByUUIDRequestBodyOp:
    if value in UPDATE_SERVICE_GROUP_BY_UUID_REQUEST_BODY_OP_VALUES:
        return cast(UpdateServiceGroupByUUIDRequestBodyOp, value)
    raise TypeError(
        f"Unexpected value {value!r}. Expected one of {UPDATE_SERVICE_GROUP_BY_UUID_REQUEST_BODY_OP_VALUES!r}"
    )



from collections.abc import Mapping
from typing import Any, TypeVar, Union
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.response_status import ResponseStatus, check_response_status
from ..types import UNSET, Unset

T = TypeVar("T", bound="DeleteCertificatesResponseDeletedCertificate")


@_attrs_define
class DeleteCertificatesResponseDeletedCertificate:
    """Details of the certificate which was deleted by this request."""

    status: Union[Unset, ResponseStatus] = UNSET
    """ The response status of an API request. """
    uuid: Union[Unset, UUID] = UNSET
    """ The UUID of the certificate which was deleted. """
    name: Union[Unset, str] = UNSET
    """ The name of the certificate which was deleted. """
    message: Union[Unset, str] = UNSET
    """ An optional message providing additional information about the status.
    This field is useful when the status is not `success`. """
    error: Union[Unset, int] = UNSET
    """ An optional error code providing additional information about the status.
    This field is useful when the status is not `success`. """
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        status: Union[Unset, str] = UNSET
        if not isinstance(self.status, Unset):
            status = self.status

        uuid: Union[Unset, str] = UNSET
        if not isinstance(self.uuid, Unset):
            uuid = str(self.uuid)

        name = self.name

        message = self.message

        error = self.error

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if status is not UNSET:
            field_dict["status"] = status
        if uuid is not UNSET:
            field_dict["uuid"] = uuid
        if name is not UNSET:
            field_dict["name"] = name
        if message is not UNSET:
            field_dict["message"] = message
        if error is not UNSET:
            field_dict["error"] = error

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        _status = d.pop("status", UNSET)
        status: Union[Unset, ResponseStatus]
        if isinstance(_status, Unset):
            status = UNSET
        else:
            status = check_response_status(_status)

        _uuid = d.pop("uuid", UNSET)
        uuid: Union[Unset, UUID]
        if isinstance(_uuid, Unset):
            uuid = UNSET
        else:
            uuid = UUID(_uuid)

        name = d.pop("name", UNSET)

        message = d.pop("message", UNSET)

        error = d.pop("error", UNSET)

        delete_certificates_response_deleted_certificate = cls(
            status=status,
            uuid=uuid,
            name=name,
            message=message,
            error=error,
        )

        delete_certificates_response_deleted_certificate.additional_properties = d
        return delete_certificates_response_deleted_certificate

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Literal, cast

UpdateInstancesRequestOp = Literal["add", "del", "set"]

UPDATE_INSTANCES_REQUEST_OP_VALUES: set[UpdateInstancesRequestOp] = {
    "add",
    "del",
    "set",
}


def check_update_instances_request_op(value: str) -> UpdateInstancesRequestOp:
    if value in UPDATE_INSTANCES_REQUEST_OP_VALUES:
        return cast(UpdateInstancesRequestOp, value)
    raise TypeError(f"Unexpected value {value!r}. Expected one of {UPDATE_INSTANCES_REQUEST_OP_VALUES!r}")



from typing import Literal, cast

CreateInstanceRequestRestartPolicy = Literal["always", "never", "on_failure"]

CREATE_INSTANCE_REQUEST_RESTART_POLICY_VALUES: set[CreateInstanceRequestRestartPolicy] = {
    "always",
    "never",
    "on_failure",
}


def check_create_instance_request_restart_policy(value: str) -> CreateInstanceRequestRestartPolicy:
    if value in CREATE_INSTANCE_REQUEST_RESTART_POLICY_VALUES:
        return cast(CreateInstanceRequestRestartPolicy, value)
    raise TypeError(f"Unexpected value {value!r}. Expected one of {CREATE_INSTANCE_REQUEST_RESTART_POLICY_VALUES!r}")
